// file: dib.cpp
// needs: windows.h, malloc.h, msgbox.h, dib.h, dib.cpp
// Interface for windows DIB support.
// Environment: Windows 3.11 + Win32s, Windows NT, Windows 95, UNIX may be
// Portability: Low
// Author: G.Pulcini
// Revisions: 
//      1.5 -  2 Oct 1995 - 32bit Flat Memory Model support
//      1.3 -  7 Feb 1995

// NT: removing huge

#ifndef __DIB_H__
#include "dib.h"
#endif

#include "msgbox.h"

DWORD GetDibInfoHeaderSize(BYTE  * lpDib)
{  return ((BITMAPINFOHEADER  *) lpDib)->biSize; }

WORD GetDibWidth(BYTE  * lpDib)
{
  if (GetDibInfoHeaderSize(lpDib) == sizeof(BITMAPCOREHEADER))
      return (WORD) (((BITMAPCOREHEADER  *) lpDib)->bcWidth);
  else
      return (WORD) (((BITMAPINFOHEADER huge *) lpDib)->biWidth);
}

WORD GetDibHeight(BYTE  * lpDib)
{
  if (GetDibInfoHeaderSize(lpDib) == sizeof(BITMAPCOREHEADER))
      return (WORD) (((BITMAPCOREHEADER  *) lpDib)->bcHeight);
  else
      return (WORD) (((BITMAPINFOHEADER  *) lpDib)->biHeight);
}

WORD GetDibBitCount(BYTE  * lpDib)
{
  WORD wBitCount;

  if (GetDibInfoHeaderSize(lpDib) == sizeof(BITMAPCOREHEADER))
      wBitCount = ((BITMAPCOREHEADER  *) lpDib)->bcBitCount;
  else
      wBitCount = ((BITMAPINFOHEADER  *) lpDib)->biBitCount;

  return wBitCount;
}

DWORD GetDibCompression(BYTE  * lpDib)
{
  DWORD dwBitCompression;

  if (GetDibInfoHeaderSize(lpDib) == sizeof(BITMAPCOREHEADER))
      dwBitCompression = 0;
  else
      dwBitCompression = ((BITMAPINFOHEADER  *) lpDib)->biCompression;

  return dwBitCompression;
}


int GetDibType(BYTE  * lpDib)
/*   1 : CORE OS/2 2.0
     0 : INFO Windows  */
{
  return (GetDibInfoHeaderSize(lpDib) == sizeof(BITMAPCOREHEADER));
}

WORD GetDibBitPlanes(BYTE  * lpDib)
{
  WORD wBitPlanes;

  if (GetDibInfoHeaderSize(lpDib) == sizeof(BITMAPCOREHEADER))
      wBitPlanes = ((BITMAPCOREHEADER  *) lpDib)->bcPlanes;
  else
      wBitPlanes = ((BITMAPINFOHEADER  *) lpDib)->biPlanes;

  return wBitPlanes;
}

BYTE * GetDibBitsAddr(BYTE  * lpDib)
{
  DWORD dwNumColors, dwColorTableSize;
  WORD  wBitCount;

  if (GetDibInfoHeaderSize(lpDib) == sizeof(BITMAPCOREHEADER))
    {
      wBitCount = ((BITMAPCOREHEADER  *) lpDib)->bcBitCount;
      if (wBitCount !=24)
        dwNumColors = 1L << wBitCount;
      else
        dwNumColors = 0;
      dwColorTableSize = dwNumColors * sizeof(RGBTRIPLE);
    }
  else
    {
      wBitCount = ((BITMAPINFOHEADER  *) lpDib)->biBitCount;
      if (GetDibInfoHeaderSize(lpDib) >= 36)
        dwNumColors = ((BITMAPINFOHEADER  *) lpDib)->biClrUsed;
      else
        dwNumColors = 0;

      if (dwNumColors == 0)
        {
          if (wBitCount != 24)
            dwNumColors = 1L << wBitCount;
          else
            dwNumColors = 0;
        }
      dwColorTableSize = dwNumColors * sizeof(RGBQUAD);
    }

  return lpDib + GetDibInfoHeaderSize(lpDib) + dwColorTableSize;
}

BYTE * ReadDib(char * szFileName)
/*
  Read a DIB file (.BMP) named szFileName and load it into a Global-Memory-Segment.
  The allocation is made through WIN API GlobalAlloc call.
  If something go wrong during loading, it returns NULL.
*/
{
  BITMAPFILEHEADER bmfh;
  BYTE * lpDib;
  DWORD dwDibSize, dwOffset, dwHeaderSize;
  int hFile;
  WORD wDibRead;

  if (-1 == (hFile = _lopen(szFileName, OF_READ | OF_SHARE_DENY_WRITE)))
    return NULL;

  if (_lread(hFile, (LPSTR) &bmfh, sizeof(BITMAPFILEHEADER)) !=
                                   sizeof(BITMAPFILEHEADER))
    {
      _lclose(hFile);
      return NULL;
    }

  if (bmfh.bfType != * (WORD *) "BM")
    {
      _lclose(hFile);
      return NULL;
    }

  dwDibSize = bmfh.bfSize - sizeof(BITMAPFILEHEADER);

  //lpDib = (BYTE huge *) GlobalAllocPtr(GMEM_MOVEABLE, dwDibSize);
  lpDib = (BYTE *) malloc(dwDibSize); //NT:

  if (lpDib == NULL)
    {
      _lclose(hFile);
      return NULL;
    }

  dwOffset = 0;
  while (dwDibSize > 0)
    {
      wDibRead = (WORD) MIN(32768ul, dwDibSize);
      if (wDibRead != _lread(hFile, (LPSTR) (lpDib + dwOffset), wDibRead))
        {
          _lclose(hFile);
          // GlobalFreePtr(lpDib);
          free(lpDib);           // NT:
          return NULL;
        }
      dwDibSize -= wDibRead;
      dwOffset += wDibRead;
    }
  _lclose(hFile);

  dwHeaderSize = GetDibInfoHeaderSize(lpDib);


  if (dwHeaderSize < 12 || (dwHeaderSize > 12 && dwHeaderSize < 16))
    {
      // GlobalFreePtr(lpDib);
      free(lpDib);
      return NULL;
    }

  return lpDib;
}

int WriteDib(char * szFileName, BYTE  * lpDib)
/*
  Writes a DIB file (.BMP) named szFileName from a Global-Memory-Segment.
  The format is INFO not CORE
  Returns TRUE if successfull
*/
{
  BITMAPFILEHEADER bmfh;
  DWORD dwDibSize, dwOffset, dwHeaderSize;
  int hFile;
  WORD wDibWrite;

  WORD wBitCount = ((BITMAPINFOHEADER  *) lpDib)->biBitCount;
  DWORD dwNumColors, dwColorTableSize;
  if (GetDibInfoHeaderSize(lpDib) >= 36)
    dwNumColors = ((BITMAPINFOHEADER  *) lpDib)->biClrUsed;
  else
    dwNumColors = 0;
  if (dwNumColors == 0)
    {
    if (wBitCount != 24)
       dwNumColors = 1L << wBitCount;
    else
       dwNumColors = 0;
    }
  dwColorTableSize = dwNumColors * sizeof(RGBQUAD);

  dwDibSize = GetDibInfoHeaderSize(lpDib)+dwColorTableSize+((BITMAPINFOHEADER*)lpDib)->biSizeImage;

  //MsgBox("", szFileName);
  if (-1 == (hFile = _lcreat(szFileName, 0)))
    return -1;
  memset(&bmfh, 0x0, sizeof(BITMAPFILEHEADER));
  //bmfh.bfType = 'BM';
  bmfh.bfType = 0x4D42;
  bmfh.bfSize = dwDibSize+sizeof(BITMAPFILEHEADER);
  bmfh.bfOffBits =sizeof(BITMAPFILEHEADER)+GetDibInfoHeaderSize(lpDib)+dwColorTableSize;

  if (_lwrite(hFile, (LPSTR) &bmfh, sizeof(BITMAPFILEHEADER)) !=
                                   sizeof(BITMAPFILEHEADER))
    {
      _lclose(hFile);
      return -2;
    }

  if (lpDib == NULL)
    {
      _lclose(hFile);
      return -3;
    }

  dwOffset = 0;
  while (dwDibSize > 0)
    {
      wDibWrite = (WORD) MIN(32768ul, dwDibSize);
      if (wDibWrite != _lwrite(hFile, (LPSTR) (lpDib + dwOffset), wDibWrite))
        {
          _lclose(hFile);
          return -4;
        }
      dwDibSize -= wDibWrite;
      dwOffset += wDibWrite;
    }
  _lclose(hFile);

/*
  dwHeaderSize = GetDibInfoHeaderSize(lpDib);

  if (dwHeaderSize < 12 || (dwHeaderSize > 12 && dwHeaderSize < 16))
    {
      GlobalFreePtr(lpDib);
      return NULL;
    }
*/

  return 0; // it's ok
}


BOOL CheckGrayScaleUncompressedDib(BYTE  * lpDib)
{
  return( GetDibBitCount(lpDib)==8 && GetDibBitPlanes(lpDib)==1
          && GetDibType(lpDib)==1 && GetDibCompression(lpDib)==0);
  // check palette to see if it is a true gray-scaled image or not.
}

// My implementation

BYTE  * Bit2RGBDib(WORD W, WORD H, BYTE  * Bits)
/*
  Makes a 24bits RGB DIB of size WxH coping Bits from another
  DIB bits pointer.
  *Bits must be in DIB format, with proper alignment.
  This function is not very fast, so use it only for small DIBs
*/
{
  BYTE  * lpDib;
  DWORD L = (W/4+((W%4) ? 1 : 0))*4;
  DWORD dwDibSize = sizeof(BITMAPINFOHEADER)+((DWORD)L)*H*3;
  // lpDib = (BYTE  *) GlobalAllocPtr(GMEM_MOVEABLE, dwDibSize);
  lpDib = (BYTE *) malloc(dwDibSize);
  ((BITMAPINFOHEADER*)lpDib)->biHeight=H;
  ((BITMAPINFOHEADER*)lpDib)->biWidth=W;
  ((BITMAPINFOHEADER*)lpDib)->biPlanes=1;
  ((BITMAPINFOHEADER*)lpDib)->biBitCount=24;
  ((BITMAPINFOHEADER*)lpDib)->biCompression=0;
  ((BITMAPINFOHEADER*)lpDib)->biSize=sizeof(BITMAPINFOHEADER);//40
  ((BITMAPINFOHEADER*)lpDib)->biSizeImage = dwDibSize-sizeof(BITMAPINFOHEADER);
  ((BITMAPINFOHEADER*)lpDib)->biXPelsPerMeter=0;
  ((BITMAPINFOHEADER*)lpDib)->biYPelsPerMeter=0;
  ((BITMAPINFOHEADER*)lpDib)->biClrUsed=0;
  ((BITMAPINFOHEADER*)lpDib)->biClrImportant=0;
  BYTE  * lpDibBits = GetDibBitsAddr(lpDib);
  //WORD cxDib = GetDibWidth(lpDib);
  //WORD cyDib = GetDibHeight(lpDib);
  DWORD I,P;
  for (WORD i=0; I=L*i,i<H; i++)
    for (WORD j=0; j<W; j++)
      {
      P=(I+j)*3;
      lpDibBits[P]=Bits[P];
      lpDibBits[P+1]=Bits[P+1];
      lpDibBits[P+2]=Bits[P+2];
      }
  return lpDib;
}

BYTE  * Bit2GrayDib(WORD W, WORD H, BYTE  * Bits)
/*
  Makes a 8bits Gray-Scale palette DIB of size WxH coping Bits from another
  DIB bits pointer.
  Set the Palette to Gray-Scale with [0]<->RGB(0,0,0) and [255]<->RGB(255,255,255)
  *Bits must be in DIB format, with proper alignment.
  This function is not very fast, so use it only for small DIBs
*/
{
  BYTE * lpDib;
  WORD L = (W/4+((W%4) ? 1 : 0))*4;
  DWORD dwDibSize = sizeof(BITMAPINFOHEADER)+((DWORD)L)*H*3+256*4;
  // lpDib = (BYTE  *) GlobalAllocPtr(GMEM_MOVEABLE, dwDibSize);
  lpDib = (BYTE*) malloc(dwDibSize);
  ((BITMAPINFOHEADER*)lpDib)->biHeight=H;
  ((BITMAPINFOHEADER*)lpDib)->biWidth=W;
  ((BITMAPINFOHEADER*)lpDib)->biPlanes=1;
  ((BITMAPINFOHEADER*)lpDib)->biBitCount=8;
  ((BITMAPINFOHEADER*)lpDib)->biCompression=0;
  ((BITMAPINFOHEADER*)lpDib)->biSize=sizeof(BITMAPINFOHEADER);//40
  ((BITMAPINFOHEADER*)lpDib)->biSizeImage = dwDibSize-sizeof(BITMAPINFOHEADER)-256*4;
  ((BITMAPINFOHEADER*)lpDib)->biXPelsPerMeter=0;
  ((BITMAPINFOHEADER*)lpDib)->biYPelsPerMeter=0;
  ((BITMAPINFOHEADER*)lpDib)->biClrUsed=0;
  ((BITMAPINFOHEADER*)lpDib)->biClrImportant=0;
  //Generates a Gray-scale palette
  for (WORD k=0; k<256; k++)
    {
      lpDib[sizeof(BITMAPINFOHEADER)+k*4+0] = k;
      lpDib[sizeof(BITMAPINFOHEADER)+k*4+1] = k;
      lpDib[sizeof(BITMAPINFOHEADER)+k*4+2] = k;
      lpDib[sizeof(BITMAPINFOHEADER)+k*4+3] = 0;
    }
  if (Bits==NULL) return lpDib;
  BYTE * lpDibBits = GetDibBitsAddr(lpDib);
  //WORD cxDib = GetDibWidth(lpDib);
  //WORD cyDib = GetDibHeight(lpDib);
  for (WORD sI,dI,i=0; dI=L*i,sI=W*i,i<H; i++)
    for (WORD j=0; j<W; j++) lpDibBits[dI+j]=Bits[sI+j];
  return lpDib;
}

BYTE  * SetGrayDibBits(BYTE  * lpDib, BYTE * Bits)
/*
  This function set Gray-Scale or paletted DIB bits to the values
  pointed by Bits parameter.
  While Bits points to contiguous bytes, they are copied in the
  right alignment for DIB.
  This function is not very fast, so use it only for small DIBs
*/
{
  BYTE  * lpDibBits = GetDibBitsAddr(lpDib);
  WORD W = GetDibWidth(lpDib);
  WORD H = GetDibHeight(lpDib);
  DWORD sI,dI, L = (W/4+((W%4) ? 1 : 0))*4;
  for (WORD i=0; dI=L*i,sI=W*i,i<H; i++)
    for (WORD j=0; j<W; j++) lpDibBits[dI+j]=Bits[sI+j];
  return lpDibBits;
}

BYTE  * CopyDib(BYTE  * lpDib)
// This allocates memory and copies a DIB into new alloc'ed area.
// Probably a faster version should be programmed in assembly using
// 32bit instruction of 386 processors.
{
  BYTE  * lpDibcpy;
  if (!lpDib) return NULL;
  DWORD dwSize = ((BITMAPINFOHEADER*)lpDib)->biSizeImage + ((BITMAPINFOHEADER*)lpDib)->biSize;
  WORD wBitCount = ((BITMAPINFOHEADER  *) lpDib)->biBitCount;
  DWORD dwNumColors, dwColorTableSize;
  if (GetDibInfoHeaderSize(lpDib) >= 36)
    dwNumColors = ((BITMAPINFOHEADER  *) lpDib)->biClrUsed;
  else
    dwNumColors = 0;
  if (dwNumColors == 0)
    {
    if (wBitCount != 24)
       dwNumColors = 1L << wBitCount;
    else
       dwNumColors = 0;
    }
  dwColorTableSize = dwNumColors * sizeof(RGBQUAD);

  dwSize = GetDibInfoHeaderSize(lpDib)+dwColorTableSize+((BITMAPINFOHEADER*)lpDib)->biSizeImage;
  // lpDibcpy = (BYTE huge *) GlobalAllocPtr(GMEM_MOVEABLE, dwSize);
  lpDibcpy = (BYTE *) malloc(dwSize);

  if (!lpDibcpy) return NULL;

  //Copy
  DWORD dwOffset = 0;
  DWORD dwDibSize = dwSize;
  while (dwDibSize > 0)
   {
   if (dwDibSize >= 0x8000) memcpy((LPSTR)(lpDibcpy+dwOffset), (LPSTR)(lpDib+dwOffset), 0x8000);
   else memcpy((LPSTR)(lpDibcpy+dwOffset), (LPSTR)(lpDib+dwOffset), dwDibSize);
   if (dwDibSize >= 0x8000) dwDibSize -= 0x8000;
   else dwDibSize = 0;
   dwOffset += 0x8000;
   }
  return lpDibcpy;
}

BYTE * CreateRGBDib(WORD W, WORD H, BYTE R, BYTE G, BYTE B)
/*
  Makes a 24bits RGB DIB of size WxH zeroing Bits.
*/
{
  BYTE  * lpDib;
  DWORD L = (W/4+((W%4) ? 1 : 0))*4;
  DWORD dwDibSize = sizeof(BITMAPINFOHEADER)+((DWORD)L)*H*3;
  lpDib = (BYTE *) malloc(dwDibSize);
  ((BITMAPINFOHEADER*)lpDib)->biHeight=H;
  ((BITMAPINFOHEADER*)lpDib)->biWidth=W;
  ((BITMAPINFOHEADER*)lpDib)->biPlanes=1;
  ((BITMAPINFOHEADER*)lpDib)->biBitCount=24;
  ((BITMAPINFOHEADER*)lpDib)->biCompression=0;
  ((BITMAPINFOHEADER*)lpDib)->biSize=sizeof(BITMAPINFOHEADER);//40
  ((BITMAPINFOHEADER*)lpDib)->biSizeImage = dwDibSize-sizeof(BITMAPINFOHEADER);
  ((BITMAPINFOHEADER*)lpDib)->biXPelsPerMeter=0;
  ((BITMAPINFOHEADER*)lpDib)->biYPelsPerMeter=0;
  ((BITMAPINFOHEADER*)lpDib)->biClrUsed=0;
  ((BITMAPINFOHEADER*)lpDib)->biClrImportant=0;
  BYTE  * lpDibBits = GetDibBitsAddr(lpDib);
  //WORD cxDib = GetDibWidth(lpDib);
  //WORD cyDib = GetDibHeight(lpDib);
  DWORD I,P;
  for (WORD i=0; I=L*i,i<H; i++)
    for (WORD j=0; j<W; j++)
      {
      P=(I+j)*3;
      lpDibBits[P]=R;
      lpDibBits[P+1]=G;
      lpDibBits[P+2]=B;
      }
  return lpDib;
}

