// file: dib.h
// needs: windows.h, malloc.h, msgbox.h, dib.h, dib.cpp
// Interface for windows dib support.
// Environment: Windows 3.11 + Win32s, Windows NT, Windows 95, UNIX may be
// Portability: Low
// Author: G.Pulcini
// Revisions: 
//      1.5 -  2 Oct 1995 - 32bit Flat Memory Model support 
//      1.3 -  7 Feb 1995

/*

    This set of functions allows you to load a Windows DIB into
  a windows global memory, the allocation is made on the fly and
  is never released. User must take care of freeing it when image
  is no more needed.

  Full support is provided for Windows BITMAPINFOHEADER structures
  and only partial support for Windows BITMAPCOREHEADER.

*/

// NT: removing huge pointers

#ifndef __DIB_H__
#define __DIB_H__

#include <malloc.h>

#ifndef __WINDOWS_H
#include <windows.h>
#endif

#ifndef MIN
#define MIN(a,b) ((a)<(b) ? (a) : (b))
#endif

extern DWORD GetDibInfoHeaderSize(BYTE * lpDib); // returns biSize of INFO
extern WORD  GetDibWidth(BYTE  * lpDib);          // returns Width of INFO/CORE
extern WORD  GetDibHeight(BYTE  * lpDib);         // returns Height of INFO/CORE
extern WORD  GetDibBitCount(BYTE  * lpDib);       // returns BitCount of INFO/CORE
extern DWORD GetDibCompression(BYTE  * lpDib);    // returns Compression of INFO, 0 for CORE
extern int   GetDibType(BYTE  * lpDib);           // returns 0 for INFO (Windows), 1 for CORE (OS/2 2.0)
extern WORD  GetDibBitPlanes(BYTE  * lpDib);      // returns BitPlanes for INFO/CORE
extern BOOL  CheckGrayScaleUncompressedDib(BYTE * lpDib); // check if 8bit/pixel, uncompressed, planes=1, type CORE?

extern BYTE  * GetDibBitsAddr(BYTE  * lpDib); // returns Address of bits INFO/CORE
extern BYTE  * ReadDib(char * szFileName);        // read a DIB INFO/CORE
extern int  WriteDib(char * szFileName, BYTE  * lpDib); // write a DIB INFO/CORE

// My extension, not so general
extern BYTE * Bit2RGBDib(WORD W, WORD H, BYTE  * Bits);  // from bits makes an RGB DIB (INFO)
extern BYTE * Bit2GrayDib(WORD W, WORD H, BYTE  * Bits); // from bits makes a Gray-Scale DIB (INFO)
extern BYTE * SetGrayDibBits(BYTE  * lpDib, BYTE  * Bits);  // set RAW bits to a Gray-Scale DIB (INFO)
extern BYTE * CopyDib(BYTE  * lpDib); // makes a copy of a DIB
extern BYTE * CreateRGBDib(WORD W, WORD H, BYTE R, BYTE G, BYTE B); // makes a new RGB DIB

#endif
