// file: ifsafi.c
// needs: kernel.*
// Intermediate layer in interfacing NT to IFSAF kernel.
// Author: Giovambattista Pulcini
// Written: 1995.10.1
// Revised:
//      1.0.0 - Oct 1995 

#include "ifsafi.h"

class Kernel Kernel;
char szStartingFileName[1024];
char szStatusString[1024];

void fnsplit(char * szpne, char * szpn, char * sze)
// D:PATH\NAME.EXT -> D:PATH\NAME + .EXT
{
  int i = strlen(szpne);
  while (i>0 && szpne[i]!='.') i--;
  if (szpne[i]=='.') 
    {
      strncpy(szpn, szpne, i);
      szpn[i]='\0';
      strcpy(sze, &szpne[i]);
    }
  else { strcpy(szpn,szpne); sze[0]='\0'; }
}

void InitKernel()
{
  //Kernel inits:
  Kernel.Verbose = FALSE; 
  Kernel.Changes = TRUE;// true if some of below params are changed.

  Kernel.B = 4;         // Range Block size
  Kernel.D = 8;        // Domain Block size
  Kernel.MaxLevel = 1;  // Max level of recursive subdivision
  
  Kernel.did = 2;
  Kernel.djd = 2;
  Kernel.overstep = 4;    
  Kernel.diameter = 20;
  Kernel.radius = 10;

  Kernel.alphabits = 8;
  Kernel.betabits = 8;
  Kernel.polybits = 8;

  Kernel.SimAnnSize = 100;
  Kernel.ClassScheme = CS_HISTOGRAM;  // Blocks classifing shceme
  Kernel.ApproximationType = APPT_DEG2;
  Kernel.Metric = MS;

  Kernel.EdgeThreshold = 1.0;
  Kernel.MidrangeThreshold = 0.3;
  Kernel.ShadeThreshold = 0.2; 
  
  //LPDBL Basis, Weight, invG;

  Kernel.ErrorThreshold = 10.0; 
  Kernel.ErrorThresholdR = 10.0;

  Kernel.ENC_USEPOLY = FALSE;
  Kernel.ENC_POLYONLY = FALSE;
  Kernel.ENC_USEISOMETRIES = TRUE;
  Kernel.ENC_BOUNDALPHA = TRUE;
  Kernel.ENC_QUANTIZEALPHA = TRUE;
  Kernel.ENC_STRIPSHADES = FALSE;
  Kernel.ENC_QUANTIZEPOLY = TRUE;
  Kernel.ENC_OVERSTEP = TRUE;
  Kernel.ENC_OVERSTEPUSEISOMETRIES = TRUE; 
  Kernel.ENC_USEORIENTATIONSCHEME = TRUE;
  
  Kernel.DEC_SINGLESTEP = FALSE;
  Kernel.DEC_SAVESTEPS = FALSE;

  Kernel.alphamin = -1.0;
  Kernel.alphamax = -1.0;
  Kernel.betamin = -255.0;
  Kernel.betamax = 255.0;

  // BYTE IsometryClass[24][24]; // we may init this here ...
  Kernel.EncodeAlgorithm = 0;


}

void CMColorREncode(char * sourcefne, char * targetfn)
{
      DWORD Time = GetTickCount(); 
      HCURSOR OldCursor = SetCursor(LoadCursor(0, IDC_WAIT));

      TEncStats * ES = new TEncStats;
      char szFileNameBuffer[255];
      char string[255];

      TImage Collage, Image(sourcefne); // Load the source image.

      if (!Image.lpDib) 
        {  
           sprintf(string, "Can't open input file: \n%s\n", sourcefne);
           sprintf(szStatusString, "Can't open: %s", sourcefne);
           // MessageBox(NULL, string, "Fatal", MB_OK | MB_ICONSTOP | MB_TASKMODAL);
	   Kernel.ShowMessage(string);
           delete ES;
           return;
        }

      Kernel.EncodeImageCR1( *ES, targetfn, Image, Collage, NULL, NULL, NULL);
      if (Kernel.ERRORCODE) 
        { sprintf(szStatusString, "Kernel Error: %08x", Kernel.ERRORCODE); 
	  Kernel.ShowMessage(szStatusString);
	  return; 
	} 
      
      Time = GetTickCount()-Time;

      char collagefn[256];
      strcpy(collagefn, targetfn);
      strcat(collagefn, ".bmp");
      WriteDib(collagefn, Collage.lpDib);

      sprintf(string, 
            "Image encoded in %d seconds\nB:%lu P:%lu [%d%%] T:%0.2lf IFS:%lu [%d%%] E:%0.2lf %d%%",
                   Time/1000,
                   ES->dwCountBlk, ES->dwPolyBlk, int(ES->dwPolyBlk*100/ES->dwCountBlk),
                   ES->ErrorThreshold, ES->dwIFSBlk, int(ES->dwIFSBlk*100/ES->dwCountBlk),
                   sqrt(ES->GlobalError), int(sqrt(ES->GlobalError)*100/255.0));
      //MessageBox(NULL, string, "Completed", MB_OK | MB_TASKMODAL );
      Kernel.ShowMessage(string);

      sprintf(szStatusString, 
             "Enc: %ds %d:1  B:%lu T:%0.2lf E:%0.2lf %d%%",
                   Time/1000, 1, ES->dwCountBlk, ES->ErrorThreshold, 
                   sqrt(ES->GlobalError), int(sqrt(ES->GlobalError)*100/255.0));
      SetCursor(OldCursor);
      delete ES;
}

void CMColorRDecode(char * startingfne, char * sourcefn, char * targetfne)
{
      DWORD Time = GetTickCount();
      HCURSOR OldCursor = SetCursor(LoadCursor(0, IDC_WAIT));
      char string[255];

      TImage Starting(startingfne), Decoded;
      if (!Starting.lpDib) 
        {  
           if (startingfne[0]!='\0')
           {
             sprintf(string, "Can't open input file: \n%s\n", startingfne);
             sprintf(szStatusString, "Can't open: %s", startingfne);
             //MessageBox(NULL, string, "Fatal", MB_OK | MB_ICONSTOP | MB_TASKMODAL);
	     Kernel.ShowMessage(string);
             return;
           }
        }
      
      /*
      char szFileName[MAXPATH],szName[MAXFILE], szDrive[MAXDRIVE], szDir[MAXDIR], szExt[MAXEXT];
      fnsplit(szCurrentEncodedFile,szDrive,szDir,szName,szExt);
      fnmerge(szFileName, szDrive,szDir,szName,".STS");
      FILE * f;
      if (f=fopen(szFileName,"rb"))
        {
          Child->ES = new TEncStats;
          if (fread(Child->ES, sizeof(TEncStats), 1, f) != 1)
            { delete Child->ES; Child->ES = NULL; }
          fclose(f);
        }
       */
      // we should find a windows containing original image
      
      PTImage img0 = NULL; // This was the original image, to compute SNR
      
      /*
      if (FirstDecode)
        img0 = &(IMGCHILD->Image);
      else if (Child->ES)
        img0 = new TImage(ReadDib(Child->ES->szInFileName), Child->ES->szInFileName);
      */

      TDecStats * DS = new TDecStats;

      //BOOL ok = Kernel.DecodeImageCR1( DS, szCurrentEncodedFile, img0, 
      //        IMGCHILD->Image, Child->Image, DEC_STEPS, FileNameCount, FirstDecode);
      
      
      BOOL ok = Kernel.DecodeImageCR1( DS, sourcefn, img0,  
        Starting, Decoded, Kernel.decsteps, 1, FALSE);
     
      if (!ok)  
        {
	  Kernel.ShowMessage("Starting image of different size. Aborted.");
          if (MessageBox(NULL, "Starting image is of different\n"
                              "size of the one encoded in IFS file.\n"
                              "Continue anyway?",
                              "Decode Warning", MB_OKCANCEL | MB_TASKMODAL | MB_ICONQUESTION)
              != IDOK)
              {
                // if (!FirstDecode && img0) delete img0;
                // delete Child;
                SetCursor(OldCursor);
                return;
              }
          else
            // ok = Kernel.DecodeImage4R( DS, szCurrentEncodedFile, img0, IMGCHILD->Image, Child->Image, DEC_STEPS, FileNameCount, FirstDecode);
            return;
        }

      Time = GetTickCount()-Time;

      WriteDib(targetfne, Decoded.lpDib);

      sprintf(string, 
            "Image decoded in %d seconds\n",Time/1000);
      //MessageBox(NULL, string, "Completed", MB_OK | MB_TASKMODAL );
      Kernel.ShowMessage(string);
      sprintf(szStatusString, "Dec: %ds", Time/1000);

      // if (!FirstDecode && img0) delete img0;
      // FirstDecode = FALSE;

      /*
      char sz1[80];
      sprintf(sz1, "P:%u drms:%0.6lf SNR:%0.2lf",
                   DEC_STEPS, DS->deltarmsError, DS->SNR);
      strcpy(Child->Message,sz1);
      if (!FirstDecode)
      {
      sprintf(sz1, "%u", ++FileNameCount);
      WritePrivateProfileString("General", "FNCount", sz1, INI_FILE);
      }

      WORD W,H;
      Child->GetWindowDimensions(W,H);
      Child->Attr.W = W;
      Child->Attr.H = H;
      GetApplication()->MakeWindow( Child );
      */

      delete DS;
      SetCursor(OldCursor);
}

BOOL IFSAFKernelInterface(int Action, char * szInputFileName, char * szOutputFileName)
// return true if ok, false on abort
{
  char szFilePathName[_MAX_PATH];    
  char szFileExt[_MAX_EXT];

    switch(Action)
        {
          case 0: // ENCODING
	  /*
            if (MessageBox( NULL, "This operation may take few minutes,\nare you sure?", 
                        "Encoding Warning", MB_YESNO | MB_TASKMODAL | MB_ICONQUESTION )
              == IDNO) return FALSE;
	   */
            // strip extension from szOutputFilenName
            fnsplit(szOutputFileName, szFilePathName, szFileExt);
            CMColorREncode(szInputFileName, szFilePathName);
            break;

          case 1: // DECODING
	  /*
            if (MessageBox( NULL, "This operation may take few minutes,\nare you sure?", 
                        "Decoding Warning", MB_YESNO | MB_TASKMODAL | MB_ICONQUESTION )
              == IDNO) return FALSE;
	   */
            fnsplit(szInputFileName, szFilePathName, szFileExt);
            CMColorRDecode(szStartingFileName, szFilePathName, szOutputFileName);
            break;
        }
  return TRUE;
}

