// file: ifsdlglt.c
// needs: ifsafi.*
// Windows NT Dialog Box Interface to IFSAF kernel.
// Author: Giovambattista Pulcini
// Written: 1995.10.1
// Revised:    
//      1.0.4 - Oct 1995 

#include <windows.h>          
#include <commdlg.h>
#include <string.h>

#include "msgbox.h"
#include "ifsafi.h"
#include "timecur.h"

static HBRUSH hbrBackground;
#include "ifsdlgs.c"

const char szHelpFileName[] = "IFSHLP.HLP";

long PASCAL WindowProc ( HWND, UINT, UINT, LONG );

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
           LPSTR lpszCmdLine, int nCmdShow)
{
    static char szAppName[34] = "IFSDLGLT";
    HWND  hwnd;
    MSG   msg;
    WNDCLASS wc;

    //static FARPROC lpfnWndProc = MakeProcInstance((FARPROC)WindowProc, hInst);
    
    if (!hPrevInstance)
    {
    wc.style        = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc  = WindowProc;
    wc.cbClsExtra   = 0;
    wc.cbWndExtra   = DLGWINDOWEXTRA;
    wc.hInstance    = hInstance;
    wc.hIcon        = LoadIcon(hInstance, "ICON_1");
    wc.hCursor      = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE+1);
    wc.lpszMenuName = NULL;
    wc.lpszClassName= szAppName;

    if (!RegisterClass(&wc)) MessageBox(NULL, "Can't register class.", "Failed", MB_OK);
    }
    hwnd = CreateDialog(hInstance, "DIALOG_1", 0, NULL);
    if (hwnd==0) MessageBox(NULL, "Can't create DIALOG_1", "Failed", MB_OK);

    SendMessage(hwnd, WM_INITDIALOG, 0, 0L);
    ShowWindow(hwnd, nCmdShow);

    SetupCursors(hInstance);

    while (GetMessage(&msg, NULL, 0, 0))      
    {                       
    TranslateMessage(&msg);
    DispatchMessage(&msg);       
    }
    // UnregisterClass(szAppName, hInstance);
    return msg.wParam;
}

static OPENFILENAME ofn;
static OPENFILENAME sfn;
static char szInputFileName[_MAX_PATH];
static char szInputTitleName[_MAX_FNAME+_MAX_EXT];
static char szOutputFileName[_MAX_PATH];
static char szOutputTitleName[_MAX_FNAME+_MAX_EXT];
static int Action;
enum act { actENCODING, actDECODING }; // Actions 

void InitializeBrowse(HWND hwnd, OPENFILENAME & ofn, int index)
{
  static char * szFilter[] = { "BMP RGB Encoded", "*.bmp",
                   "IFS Compressed ", "*.ifs", "" };
  ofn.lStructSize       = sizeof(OPENFILENAME);
  ofn.hwndOwner         = hwnd;
  ofn.hInstance         = NULL;
  ofn.lpstrFilter       = szFilter[0];
  ofn.nFilterIndex      = index;
  ofn.lpstrCustomFilter = 0;
  ofn.nMaxCustFilter    = 0;
  ofn.lpstrFile         = NULL;
  ofn.nMaxFile          = _MAX_PATH;
  ofn.lpstrFileTitle    = NULL;
  ofn.nMaxFileTitle     = _MAX_FNAME + _MAX_EXT;
  ofn.lpstrInitialDir   = NULL;
  ofn.lpstrTitle        = NULL;
  ofn.Flags             = 0;
  ofn.nFileOffset       = 0;
  ofn.nFileExtension    = 0;
  ofn.lpstrDefExt       = "bmp";
  ofn.lCustData         = 0l;
  ofn.lpfnHook          = NULL;
  ofn.lpTemplateName    = NULL;
}

BOOL InputFileBrowse(HWND hwnd, LPSTR lpstrFileName, LPSTR lpstrTitleName)
{

    ofn.hwndOwner       = hwnd; 
    ofn.lpstrFile       = lpstrFileName;
    ofn.lpstrFileTitle  = lpstrTitleName;
    ofn.lpstrDefExt     = ((Action==actENCODING) ? "bmp" : "ifs");
    ofn.nFilterIndex    = ((Action==actENCODING) ? 1 : 2);
    ofn.Flags           = OFN_FILEMUSTEXIST;

    return GetOpenFileName( &ofn );
}

BOOL OutputFileBrowse(HWND hwnd, LPSTR lpstrFileName, LPSTR lpstrTitleName)
{
    sfn.hwndOwner       = hwnd; 
    sfn.lpstrFile       = lpstrFileName;
    sfn.lpstrFileTitle  = lpstrTitleName;
    sfn.lpstrDefExt     = ((Action==actENCODING) ? "ifs" : "bmp");
    sfn.nFilterIndex    = ((Action==actENCODING) ? 2 : 1);
    sfn.Flags           = OFN_OVERWRITEPROMPT;

    return GetSaveFileName( &sfn );
}

long PASCAL AboutDlgProc(HWND hdlg, UINT message, UINT wParam, LONG lParam)
{
    switch (message)
      {
    case WM_INITDIALOG:
        return TRUE;

    case WM_CTLCOLORDLG:
    case WM_CTLCOLORSTATIC:
    case WM_CTLCOLORBTN:
        SetBkColor((HDC)((wParam)),  GetSysColor(COLOR_BTNFACE));
        return (DWORD) hbrBackground;

    case WM_SYSCOLORCHANGE:
        DeleteObject(hbrBackground);
        hbrBackground = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
        return 0;

    case WM_COMMAND:
        switch (wParam)
          {
        case IDOK: 
            EndDialog(hdlg, 0);
            return TRUE;
          }
        break;
      }
    return FALSE;
}

inline HFONT MakeFont(HDC hdc, char* Name, int Size)
{
  LOGFONT lf;
  memset(&lf, 0x0, sizeof(LOGFONT));
  lf.lfHeight = Size;
  lf.lfWeight = 100;
  strcpy((LPSTR)lf.lfFaceName,(LPSTR)Name);
  return SelectObject(hdc, CreateFontIndirect(&lf));
}

void PaintBarInWindow(HWND hwnd, float x)
{
  HDC hdc = GetDC(hwnd);
  RECT rect,rect1;
  GetClientRect(hwnd, &rect);
  // Rectangle(hdc, rect.left, rect.top, rect.right, rect.bottom);
  rect.left++; 
  rect.top++;
  rect1 = rect;
  if (x<0) x=0; 
  if (x>1) x=1;
  rect.right = rect.left + (rect.right-rect.left)*x; // -1 if not null ?
  // rect.bottom; // -- if pen not null
  rect1.left = rect.right;
  HPEN hp = SelectObject(hdc, GetStockObject(NULL_PEN));
  Rectangle(hdc, rect1.left, rect1.top, rect1.right, rect1.bottom);
  HBRUSH hb = SelectObject(hdc, CreateSolidBrush(RGB(0,0,128)));
  // HPEN hp = SelectObject(hdc, GetStockObject(NULL_PEN));
  Rectangle(hdc, rect.left, rect.top, rect.right, rect.bottom);
  SelectObject(hdc, hp);
  DeleteObject(SelectObject(hdc, hb));
  ReleaseDC(hwnd, hdc);
}


void PaintTextInWindow(HWND hwnd, char * message)
{
  HDC hdc = GetDC(hwnd);
  RECT rect;
  GetClientRect(hwnd, &rect);
  
  HFONT hfont = MakeFont(hdc, "MS Sans Serif", 12);
  HPEN hp = SelectObject(hdc, GetStockObject(NULL_PEN));
  HBRUSH hb = SelectObject(hdc, CreateSolidBrush(GetSysColor(COLOR_BTNFACE)));
  Rectangle(hdc, rect.left, rect.top, rect.right+1, rect.bottom+1);
  SelectObject(hdc, hp);
  SetBkColor(hdc,  GetSysColor(COLOR_BTNFACE));
  DrawText(hdc, message, -1, &rect, DT_LEFT);
  DeleteObject(SelectObject(hdc, hfont));
  DeleteObject(SelectObject(hdc, hb));
  ReleaseDC(hwnd, hdc);
}

// to communicate with DLG
static HWND PDLGtext;
static HWND PDLGbar;
void ProcessDlgProcfn(char * msg, float x)
// if msg = NULL, do not delete previous message, only show float
// if msg != NULL ignore float
// if msg[0]=0 delete the message 
{
  static char message[1024];
  static float value;
  if (msg) strcpy(message, msg); else
    if (x!=-1) value = x;

  if (msg) PaintTextInWindow(PDLGtext, message);
  if (x!=-1) PaintBarInWindow(PDLGbar, value);
  if (!msg && x == -1) // REDRAW
    {
      PaintTextInWindow(PDLGtext, message);
      PaintBarInWindow(PDLGbar, value);
    }
}

long PASCAL ProgressDlgProc(HWND hdlg, UINT message, UINT wParam, LONG lParam)
{
    static BOOL WM_IDOKposted = FALSE;

    switch (message)
      {
    case WM_INITDIALOG:
       PDLGtext = GetDlgItem(hdlg, 102); // text
       PDLGbar  = GetDlgItem(hdlg, 100); // bar
       return TRUE;

    case WM_CTLCOLORDLG:
    case WM_CTLCOLOREDIT:
    case WM_CTLCOLORSTATIC:
    case WM_CTLCOLORBTN:
        SetBkColor((HDC)((wParam)),  GetSysColor(COLOR_BTNFACE));
        return (DWORD) hbrBackground;

    case WM_SYSCOLORCHANGE:
        DeleteObject(hbrBackground);
        hbrBackground = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
        return 0;

    case WM_PAINT:
        if (!WM_IDOKposted) 
          { WM_IDOKposted = TRUE;
            PostMessage(hdlg, WM_COMMAND, (WPARAM)IDOK, 0l);
          } 
        else ProcessDlgProcfn(NULL, -1); // Redraw
        return FALSE; // windows paints

    case WM_COMMAND:
        switch (wParam)
          {
        case IDOK: 
            {
            Kernel.Processfn = ProcessDlgProcfn;
            ProcessDlgProcfn("Starting Encoding...", 0);
            
            BOOL ok = IFSAFKernelInterface(Action, szInputFileName, szOutputFileName);      
 
            Kernel.Processfn = NULL;        
            EnableWindow(GetDlgItem(hdlg, 101), TRUE);
            SetWindowText(GetDlgItem(hdlg, 101), "OK");
            ProcessDlgProcfn(NULL, (ok ? 1.0:0.0));
            if (!ok) PostMessage(hdlg, WM_COMMAND, (WPARAM)101, 0l);
            return TRUE;
            }
        case 101: // Abort
            WM_IDOKposted = FALSE;
            EndDialog(hdlg, 0);
            return TRUE;
          }
        break;
      }
    return FALSE;
}


long PASCAL WindowProc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
{
    HWND hw;

    static FARPROC lpfnAboutDlgProc;
    static FARPROC lpfnEncodingDlgProc;
    static FARPROC lpfnDecodingDlgProc;
    static FARPROC lpfnProgressDlgProc;
    static HANDLE hInst;
    static HFONT hfontDlg;
    LOGFONT lFont;
    
    switch(message)
    {
    case WM_CREATE:  //inits
      // Since this is not a true dialog, there is no default windows
      // procedure that sends a WM_INITDIALOG. Also, when WM_CREATE is 
      // called, Windows has just create the dlg windows, and is about 
      // to call al dependent windows, so, here, now, they do not exist
      // yet.
      hInst = ((LPCREATESTRUCT) lParam)->hInstance;
      lpfnAboutDlgProc = MakeProcInstance((FARPROC)AboutDlgProc, hInst);
      lpfnEncodingDlgProc = MakeProcInstance((FARPROC)EncSettingsDlgProc, hInst);
      lpfnDecodingDlgProc = MakeProcInstance((FARPROC)DecSettingsDlgProc, hInst);
      lpfnProgressDlgProc = MakeProcInstance((FARPROC)ProgressDlgProc, hInst);
      InitializeBrowse(hwnd, ofn, 0); // BMP
      InitializeBrowse(hwnd, sfn, 1); // IFS
      Action = actENCODING;
      // Default FNs are NULL
      hbrBackground = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
      DragAcceptFiles(hwnd, TRUE);
      return 0;
                          
    case WM_INITDIALOG:       
      // We simulate this message from main...
      // Windows does not actually send this to this wndproc.

      SendDlgItemMessage(hwnd, 110, BM_SETCHECK, 1, 0l); // Encode
      SendDlgItemMessage(hwnd, 111, BM_SETCHECK, 0, 0l); // Decode
      
      hfontDlg = GetStockObject(ANSI_VAR_FONT);
        if (GetObject(hfontDlg, sizeof(LOGFONT), (LPSTR) &lFont)) {
          lFont.lfWeight = FW_NORMAL;
          if (hfontDlg = CreateFontIndirect((LPLOGFONT) &lFont)) {
            SendDlgItemMessage(hwnd, 120, WM_SETFONT,(WPARAM) hfontDlg, 0); // About
            SendDlgItemMessage(hwnd, 110, WM_SETFONT,(WPARAM) hfontDlg, 0); // Encoding (.)
            SendDlgItemMessage(hwnd, 111, WM_SETFONT,(WPARAM) hfontDlg, 0); // Decoding (.)
            SendDlgItemMessage(hwnd, 112, WM_SETFONT,(WPARAM) hfontDlg, 0); // Browse Input
            SendDlgItemMessage(hwnd, 113, WM_SETFONT,(WPARAM) hfontDlg, 0); // Input Edit
            SendDlgItemMessage(hwnd, 114, WM_SETFONT,(WPARAM) hfontDlg, 0); // Browse Output
            SendDlgItemMessage(hwnd, 115, WM_SETFONT,(WPARAM) hfontDlg, 0); // Output Edit
            SendDlgItemMessage(hwnd, 117, WM_SETFONT,(WPARAM) hfontDlg, 0); // Settings
            SendDlgItemMessage(hwnd, 118, WM_SETFONT,(WPARAM) hfontDlg, 0); // Help
            SendDlgItemMessage(hwnd, 119, WM_SETFONT,(WPARAM) hfontDlg, 0); // Quit
            SendDlgItemMessage(hwnd, 130, WM_SETFONT,(WPARAM) hfontDlg, 0); // Status string
            }
           }
       return 0;

    case WM_CTLCOLORDLG:
    case WM_CTLCOLORSTATIC:
    case WM_CTLCOLORBTN:
        SetBkColor((HDC)((wParam)),  GetSysColor(COLOR_BTNFACE));
        //SetBkMode((HDC)((wParam)), TRANSPARENT);
        return (DWORD) hbrBackground;

    case WM_SYSCOLORCHANGE:
        DeleteObject(hbrBackground);
        hbrBackground = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
        return 0;

    case WM_COMMAND:      
        //SetFocus(hwnd);
        switch( LOWORD(wParam) )  {
          
          case 120: // About
            DialogBox(hInst, "DIALOG_3", hwnd, lpfnAboutDlgProc);
            return 0;

          case 110: // (o) Encode
            if (Action==actDECODING) {
                GetWindowText(GetDlgItem(hwnd,113), szInputFileName, _MAX_PATH);
                GetWindowText(GetDlgItem(hwnd,115), szOutputFileName, _MAX_PATH);
                SetWindowText(GetDlgItem(hwnd,113), szOutputFileName);
                SetWindowText(GetDlgItem(hwnd,115), szInputFileName);
                }
            Action = actENCODING;
            break;

          case 111: // (o) Decode
            if (Action==actENCODING) {
                GetWindowText(GetDlgItem(hwnd,113), szInputFileName, _MAX_PATH);
                GetWindowText(GetDlgItem(hwnd,115), szOutputFileName, _MAX_PATH);
                SetWindowText(GetDlgItem(hwnd,113), szOutputFileName);
                SetWindowText(GetDlgItem(hwnd,115), szInputFileName);
                }
            Action = actDECODING;
            break;

          case 112: // Browse Input - Open File
            GetWindowText(GetDlgItem(hwnd,113), szInputFileName, _MAX_PATH);
            szInputTitleName[0] = '\0';
            if (SendDlgItemMessage(hwnd, 110, BM_GETCHECK, 0, 0l))
                Action = actENCODING; else Action = actDECODING;
            if (InputFileBrowse(hwnd, szInputFileName, szInputTitleName))
                SetWindowText(GetDlgItem(hwnd, 113), szInputFileName);
            //SendDlgItemMessage(hwnd, 113, EM_LINESCROLL,-_MAX_PATH,0l);
            return 0;     
             
          case 114: // Browse Output   
            GetWindowText(GetDlgItem(hwnd,115), szOutputFileName, _MAX_PATH);
            szOutputTitleName[0] = '\0';
            if (SendDlgItemMessage(hwnd, 110, BM_GETCHECK, 0, 0l))
                Action = actENCODING; else Action = actDECODING;
            if (OutputFileBrowse(hwnd, szOutputFileName, szOutputTitleName))
                SetWindowText(GetDlgItem(hwnd, 115), szOutputFileName);
            return 0;

          case 116: // Start 
            GetWindowText(GetDlgItem(hwnd,113), szInputFileName, _MAX_PATH);
            GetWindowText(GetDlgItem(hwnd,115), szOutputFileName, _MAX_PATH);
            if (szInputFileName[0]==0) {
                MessageBox(hwnd, "You must specify an Input file,\n"
                                 "by hand, by clicking on browse or\n"
                                 "by dragging and dropping.", "Error",
                                 MB_OK|MB_TASKMODAL|MB_ICONEXCLAMATION);
                return 0;
                }
            if (szOutputFileName[0]==0) {
                if (Action == actENCODING)
                    strcpy(szOutputFileName, "out.ifs");
                else
                    strcpy(szOutputFileName, "dec.bmp");
                SetWindowText(GetDlgItem(hwnd, 115), szOutputFileName);
                }
            DialogBox(hInst, "DIALOG_2", hwnd, lpfnProgressDlgProc);
            SetWindowText(GetDlgItem(hwnd,130), szStatusString);
            return 0;  

          case 117: // Settings
            if (Action == actENCODING)
              DialogBox(hInst, "DIALOG_5", hwnd, lpfnEncodingDlgProc);
            else
              DialogBox(hInst, "DIALOG_4", hwnd, lpfnDecodingDlgProc);
            return 0;

          case 118: // Help
            WinHelp(hwnd,szHelpFileName,HELP_INDEX,0L);
            return 0;
          
          case 119:  // Quit
            SendMessage(hwnd, WM_CLOSE, 0, 0L);
            return 0;
          }
        return 0;

    case WM_DROPFILES:
        if (DragQueryFile((HANDLE)wParam, -1, (LPSTR)NULL, 0) > 1)
            MessageBox(hwnd, "Too many files selected:\nonly the first file will be used.", 
                    "Warning", MB_TASKMODAL|MB_ICONEXCLAMATION|MB_OK);
        DragQueryFile((HANDLE)wParam, 0, szInputFileName, _MAX_PATH);
        szInputTitleName[0] = '\0';
        if (SendDlgItemMessage(hwnd, 110, BM_GETCHECK, 0, 0l))
            Action = actENCODING; else Action = actDECODING;
        SetWindowText(GetDlgItem(hwnd, 113), szInputFileName);
        DragFinish((HANDLE)wParam);
        return 0;

    case WM_DESTROY:
        ReleaseCursors();
        DeleteObject(hbrBackground);
        PostQuitMessage(0);
        return 0;
    }
    return DefWindowProc(hwnd, message, wParam, lParam);
}

