// file:  image.cpp
// needs: dib.h, dib.cpp, krnldef.h
// Windows DIB image class
// Written by G.Pulcini
// Last revised:
//      1.6 - Oct 1995 - 32bit Flat Memory Model support (NT)
//      1.5 - Jan 1995

#ifndef __IMAGE_H__
#include "image.h"
#endif

#ifndef __STRING_H
#include <string.h>
#endif

/////////////////////////////////////////////////////////////////////
// TImage

TImage::TImage(LPBYTE lpDib0, char * fn)
{
  Flag = 0;
  lpDib = lpDib0;
  if (lpDib)
    {
      wWidth = GetDibWidth(lpDib);
      wHeight= GetDibHeight(lpDib);
      bRGB = GetDibBitCount(lpDib)==24;
      if (bRGB) dwLINESIZE = ((wWidth*3/4)+(((wWidth*3)%4) ? 1:0))*4;
      else      dwLINESIZE = ((wWidth/4) + ((wWidth%4) ? 1 : 0))*4;
      lpBitMap = GetDibBitsAddr(lpDib);
    }
  else lpBitMap = NULL;
  if (fn) strcpy(szFileName, fn); else szFileName[0]=0;
}

TImage::TImage(char * fn)
{
  Flag = 0;
  strcpy(szFileName, fn);
  lpDib = ReadDib(szFileName);
  if (lpDib)
    {
      wWidth = GetDibWidth(lpDib);
      wHeight= GetDibHeight(lpDib);
      bRGB = GetDibBitCount(lpDib)==24;
      if (bRGB) dwLINESIZE = ((wWidth*3/4)+(((wWidth*3)%4) ? 1:0))*4;
      else      dwLINESIZE = ((wWidth/4) + ((wWidth%4) ? 1 : 0))*4;
      lpBitMap = GetDibBitsAddr(lpDib);
    }
  else lpBitMap = NULL, szFileName[0]=0;
}


TImage::TImage(const TImage & img)
{
  Flag = 0;
  wWidth = img.wWidth;
  wHeight= img.wHeight;
  bRGB = img.bRGB;
  dwLINESIZE = img.dwLINESIZE;
  lpDib = CopyDib(img.lpDib);
  lpBitMap = GetDibBitsAddr(lpDib);
}

void TImage::Copy(const RTImage img)
{
  if (lpDib && !(Flag & IMG_NODISCARD)) FREE(lpDib);
  Flag = 0;
  wWidth = img.wWidth;
  wHeight= img.wHeight;
  bRGB = img.bRGB;
  dwLINESIZE = img.dwLINESIZE;
  lpDib = CopyDib(img.lpDib);
  lpBitMap = GetDibBitsAddr(lpDib);
}

TImage::~TImage()
{
  if (lpDib && !(Flag & IMG_NODISCARD)) FREE(lpDib);
}

void TImage::Setup(LPBYTE lpDib0, char * fn)
{
  if (lpDib && !(Flag & IMG_NODISCARD)) FREE(lpDib);
  Flag = 0;
  lpDib = lpDib0;
  if (lpDib)
    {
      wWidth = GetDibWidth(lpDib);
      wHeight= GetDibHeight(lpDib);
      bRGB = GetDibBitCount(lpDib)==24;
      if (bRGB) dwLINESIZE = ((wWidth*3/4)+(((wWidth*3)%4) ? 1:0))*4;
      else      dwLINESIZE = ((wWidth/4) + ((wWidth%4) ? 1 : 0))*4;
      lpBitMap = GetDibBitsAddr(lpDib);
    }
  else lpBitMap = NULL;
  if (fn) strcpy(szFileName, fn); else szFileName[0]=0;
}

void TImage::GetBlock(DWORD i0, DWORD j0, DWORD N, LPBYTE Blk) const
// (0,0) in image is lower left corner. (0,0) in block is lower left corner too.
// If you interpret block as a matrix with (0,0) beeing upper left corner, you flip
// image vertically.
{
  DWORD sI,dI,J;
  if (bRGB)
    for(WORD i=0; sI=dwLINESIZE*(i+i0),dI=N*3*i,i<N; i++)
      for(WORD j=0; J=j+j0*3, j<3*N; j++)
        Blk[dI+j]=lpBitMap[sI+J];
  else
    for(WORD i=0; sI=dwLINESIZE*(i+i0),dI=N*i,i<N; i++)
      for(WORD j=0; j<N; j++) Blk[dI+j]=lpBitMap[sI+j+j0];
}

void TImage::SetBlock(DWORD i0, DWORD j0, DWORD N, LPBYTE Blk)
{
  DWORD sI,dI,J;
  if (bRGB)
    if(N+i0<=wHeight && N+j0<=wWidth)
    for(WORD i=0; dI=dwLINESIZE*(i+i0),sI=N*3*i,i<N; i++)
      for(WORD j=0; J=j+j0*3, j<3*N; j++) lpBitMap[dI+J]=Blk[sI+j];
    else
    {
    WORD N1,W1;
    if(N+i0<=wHeight) N1=N; else N1=wHeight-i0;
    if(N+j0<=wWidth)  W1=N; else W1=wWidth-j0;
    for(WORD i=0; dI=dwLINESIZE*(i+i0),sI=N*3*i,i<N1; i++)
      for(WORD j=0; J=j+j0*3, j<3*W1; j++) lpBitMap[dI+J]=Blk[sI+j];
    }
  else
    for(WORD i=0; dI=dwLINESIZE*(i+i0),sI=N*i,i<N; i++)
      for(WORD j=0; j<N; j++) lpBitMap[dI+j+j0]=Blk[sI+j];
}


