// file:  image.h
// needs: dib.h, dib.cpp, krnldef.h
// Windows DIB image class
// Written by G.Pulcini
// Last revised:
//      1.6 - Oct 1995 - 32bit Flat Memory Model support (NT)
//      1.5 - Jan 1995

#ifndef __IMAGE_H__
#define __IMAGE_H__

#ifndef __DIB_H__
#include "dib.h"
#endif

#ifndef __KRNLDEF_H__
#include "krnldef.h"
#endif

// This class is specific file-format oriented. We use Windows DIB.
// However it's system environment independent

// Extra long Pointer to BYTE
// typedef BYTE huge * LPBYTE; defined in windows.h

#define LPBYTE BYTE *
#define ALLOCATE(n) malloc(n)
#define FREE(p) free(p)


#define IMG_NODISCARD 0x0001

struct TPixel { BYTE R,G,B; };

_CLASSDEF(TImage)
class TImage
{
private:
  DWORD  dwLINESIZE;
  BOOL   bRGB;

public:
  WORD Flag;
  LPBYTE lpDib;
  WORD   wWidth, wHeight;
  LPBYTE lpBitMap;
  char   szFileName[64];

  TImage(LPBYTE lpDib0 = NULL, char * fn = NULL);
  TImage(char * fn);
  TImage(const TImage & img);
  ~TImage();

  // Note that we assume array addressing: so (i,j) indicize cartesian point (j,i)
  void Setup(LPBYTE lpDib0 = NULL, char * fn = NULL);
  BYTE huge & Pixel(DWORD i, DWORD j)
    { return (lpBitMap[i*dwLINESIZE+j]); };
  TPixel huge & RGBPixel(DWORD i, DWORD j)
    { return *((TPixel *) &(lpBitMap[i*dwLINESIZE+(j<<1)+j])); };
  BYTE Pixel(DWORD i, DWORD j) const
    { return lpBitMap[i*dwLINESIZE+j]; };
  TPixel RGBPixel(DWORD i, DWORD j) const
    { return *((TPixel *) &lpBitMap[i*dwLINESIZE+(j<<1)+j]); };
  void GetBlock(DWORD i, DWORD j, DWORD N, LPBYTE Block) const; //i,j in pixel
  void SetBlock(DWORD i, DWORD j, DWORD N, LPBYTE Block); //i,j in pixel
  void Copy(const RTImage); // make current image a copy of
};


#endif
