// file: kernel-2.cpp
// Some extensions to kernel functions
// Written by G.Pulcini
// Last revised 1995/2/18  Rev. 2.0

#include "kernel.h"

void Kernel::BlockSum(const RTImage img, WORD ib0, WORD jb0, double & WSum, double & WSum2, double SSum[4])
// returns sum and sum^2 of block
{
    WORD ib=ib0*B, jb=jb0*B, I, IB2, B2=B/2;
    WSum = 0;
    WSum2= 0;
    SSum[0]=SSum[1]=SSum[2]=SSum[3]=0;
    for (WORD i=0; I=i*B, IB2=i/B2*2, i<B; i++)
      for (WORD j=0; j<B; j++)
        {
          double x = img.Pixel(ib+i,jb+j);
          double wx = Weight[I+j]*x;
          WSum += wx;
          WSum2+= wx*x;
          SSum[IB2+j/B2] += x;
        }
}

BYTE Kernel::BlockHistogram(const RTImage img, WORD ib0, WORD jb0, DWORD * Histogram)
// returns middle gray associated with block
{
    WORD ib=ib0*B, jb=jb0*B;
    DWORD max=0, mean=0;
    for (int k=0; k<256; k++) Histogram[k]=0;
    for (WORD i=0; i<B; i++)
      for (WORD j=0; j<B; j++)
        {
          DWORD x=Histogram[img.Pixel(ib+i,jb+j)]++;
          max = Max(x,max);
        }
    for (k=0; k<256; k++) mean+=Histogram[k]*k;
    return (mean/(B*B));
}

void Kernel::ExtractBlock(const RTImage img, WORD i0, WORD j0, WORD N, BYTE  * Bits)
// This function is very general but not so fast
// Use TImage::GetBlock instead for fast block getting.
{
  for(WORD i=0; i<N; i++)
    for(WORD j=0; j<N; j++)
      Bits[i*N+j]=img.Pixel(i0*N+i,j0*N+j);
}

void Kernel::AbsDifference(PTBLOCK A, PTBLOCK B, PTBLOCK E, WORD N)
// computes E = |A-B|
{
  WORD I,i,j;
  // switch
  for(i=0; I=i*N,i<N; i++)
    for(j=0; j<N; j++)
         E[I+j] = abs(int(A[I+j])-B[I+j]);
}

void Kernel::Difference(PTBLOCK A, PTBLOCK B, PTBLOCK E, WORD N)
// computes E = A-B
{
  WORD I,i,j;
  // switch
  for(i=0; I=i*N,i<N; i++)
    for(j=0; j<N; j++)
         E[I+j] = 128 + (int(A[I+j])-B[I+j])/2;
}

void Kernel::GrayLevelScale(RTImage img, int n, int d, int L)
{
DWORD Time = GetTimeTicks();
  switch (L)
  {
  case 0:
    {
      for(WORD i=0; i<img.wHeight; i++)
        for (WORD j=0; j<img.wWidth; j++)
          img.Pixel(i,j) = img.Pixel(i,j)*n/d;
    } break;
  case 1:
    {
      for(WORD i=0; i<img.wHeight; i++)
        for (WORD j=0; j<img.wWidth; j++)
          img.Pixel(i,j) = 255-((255-img.Pixel(i,j))*n/d);
    } break;
  }
Timing = GetTimeTicks()-Time;
}

// For all Grad? check range: 128+Sum/2 may be ok for mapping derivative into 0..255 range
// {{-1,-2,-1},{0,0,0},{1,2,1}} * 0..255 -> -255*4..255*4, 4 = 1+2+1. (-255*4..255*4)/4/2 : -128..128
double Kernel::Grady(BYTE * sBlock, BYTE * dBlock, WORD N)
{
  long int y, Gy = 0;
  WORD I;
  for(WORD i=1; I=N*i,i<N-1; i++)
    for(WORD j=1; j<N-1; j++)
      dBlock[I+j] =128+( y = -int(sBlock[I-N+j-1])-int(sBlock[I-N+j]<<1)-int(sBlock[I-N+j+1])
                    +int(sBlock[I+N+j-1])+int(sBlock[I+N+j]<<1)+int(sBlock[I+N+j+1]))/8,
      Gy += y;
  for(i=0; i<N; i++)
    { dBlock[i]=0; dBlock[(N-1)*N+i]=0; dBlock[i*N]=0; dBlock[i*N+N-1]=0; }
  return double(Gy)/(N*N);
}

double Kernel::Gradx(BYTE * sBlock, BYTE * dBlock, WORD N)
{
  long int x, Gx = 0;
  WORD I;
  for(WORD i=1; I=N*i,i<N-1; i++)
    for(WORD j=1; j<N-1; j++)
      dBlock[I+j] =128+( x = -int(sBlock[I-N+j-1])-int(sBlock[I+j-1]<<1)-int(sBlock[I+N+j-1])
                    +int(sBlock[I-N+j+1])+int(sBlock[I+j+1]<<1)+int(sBlock[I+N+j+1]))/8,
      Gx += x;
  for(i=0; i<N; i++)
    { dBlock[i]=0; dBlock[(N-1)*N+i]=0; dBlock[i*N]=0; dBlock[i*N+N-1]=0; }
  return double(Gx)/(N*N);
}

void Kernel::Grad(BYTE * Gx, BYTE * Gy, BYTE  * dBlock, WORD N)
{
  for(WORD I,i=0; I=N*i,i<N; i++)
    for (WORD j=0; j<N; j++)
      dBlock[I+j]=(abs(int(Gx[I+j])-128)+abs(int(Gy[I+j])-128))*2; // we need not abs()
    //dBlock[I+j]=Sqrt(Sqr(Gx[I+j])+Sqr(Gy[I+j]));
}

void Kernel::Laplacian(BYTE * sBlock, BYTE * dBlock, WORD N)
{
  WORD I;
  for(WORD i=1; I=N*i,i<N-1; i++)
    for(WORD j=1; j<N-1; j++)
      dBlock[I+j] = 128+( -int(sBlock[I-N+j])-int(sBlock[I+j-1])+int(sBlock[I+j]<<2)
                    -int(sBlock[I+j+1])-int(sBlock[I+N+j]))/8;
  for(i=0; i<N; i++)
    { dBlock[i]=0; dBlock[(N-1)*N+i]=0; dBlock[i*N]=0; dBlock[i*N+N-1]=0; }
}

void Kernel::VSSprojection(BYTE  * Block0, WORD N, double &E, double &L, double &A, double &P)
{
  double *ES, *LS, *AS, *Block;
  const double sqrt2 = sqrt(double(2.0));
  const double Normw1234 = sqrt(double(8.0));
  const double Normw56 = 2.0;
  const double Normw78 = 6.0;
  const double Normw9 = 3.0;
  //const double Normp = sqrt(double(8.0+64.0));

  ES = new double[N*N]; // Edge subspace
  LS = new double[N*N]; // Line subspace
  AS = new double[N*N]; // Average subspace
  Block=new double[N*N]; // block itself
  WORD I; for(I=0; I<N*N; I++) Block[I]=double(Block0[I])/255;
  // per velocit sostituire con z1..z9 i Block[I+j]
  E=L=A=P=0;
  for(WORD i=1; I=N*i,i<N-1; i++)
    for(WORD j=1; j<N-1; j++)
      {
      double Normz = sqrt(
              sqr(Block[I-N+j-1])+sqr(Block[I-N+j])+sqr(Block[I-N+j+1])+
              sqr(Block[I+j-1])+sqr(Block[I+j])+sqr(Block[I+j+1])+
              sqr(Block[I+N+j-1])+sqr(Block[I+N+j])+sqr(Block[I+N+j+1])
              );
      if (Normz<1e-20) { A+=1; continue; } // we see 0 as belonging to average subspace, although it belongs to all.
      // ES[I+j]
      E += sqrt(
                sqr((Block[I-N+j-1]+sqrt2*Block[I-N+j]+Block[I-N+j+1]+
                    -Block[I+N+j-1]-sqrt2*Block[I+N+j]-Block[I+N+j+1])/Normw1234) + // w1
                sqr((Block[I-N+j-1]+sqrt2*Block[I+j-1]+Block[I+N+j-1]+
                    -Block[I-N+j+1]-sqrt2*Block[I+j+1]-Block[I+N+j+1])/Normw1234) + // w2
                sqr((-Block[I-N+j]+sqrt2*Block[I-N+j+1]+
                    Block[I+j-1]-Block[I+j+1]+
                    -sqrt2*Block[I+N+j-1]+Block[I+N+j])/Normw1234) + //w3
                sqr((sqrt2*Block[I-N+j-1]-Block[I-N+j]+
                    -Block[I+j-1]+Block[I+j+1]+
                    Block[I+N+j]-sqrt2*Block[I+N+j+1])/Normw1234) // w4
                ) / Normz; // projection on Edge subspace
      L += sqrt(
                sqr((Block[I-N+j]-Block[I+j-1]-Block[I+j+1]+Block[I+N+j])/Normw56) + //w5
                sqr((-Block[I-N+j-1]+Block[I-N+j+1]+Block[I+N+j-1]-Block[I+N+j+1])/Normw56) + //w6
                sqr((Block[I-N+j-1]-2*Block[I-N+j]+Block[I-N+j+1]+
                    -2*Block[I+j-1]+4*Block[I+j]-2*Block[I+j+1]+
                    Block[I+N+j-1]-2*Block[I+N+j]+Block[I+N+j+1])/Normw78)+ // w7
                sqr((-2*Block[I-N+j-1]+Block[I-N+j]-2*Block[I-N+j+1]+
                    Block[I+j-1]+4*Block[I+j]+Block[I+j+1]+
                    -2*Block[I+N+j-1]+Block[I+N+j]-2*Block[I+N+j+1])/Normw78) // w8
                )/Normz; // projection on Line subspace
      A += abs(
                  ( Block[I-N+j-1]+Block[I-N+j]+Block[I-N+j+1]+
                    Block[I+j-1]+Block[I+j]+Block[I+j+1]+
                    Block[I+N+j-1]+Block[I+N+j]+Block[I+N+j+1] )/Normw9 // w9
                        )/Normz; // projection on Average subspace
      P += abs(
                  ( -Block[I-N+j-1]-Block[I-N+j]-Block[I-N+j+1]+
                    -Block[I+j-1]+8*Block[I+j]-Block[I+j+1]+
                    -Block[I+N+j-1]-Block[I+N+j]-Block[I+N+j+1])      // not a base vector
                        );
      }
  E /= 1.0*N*N;
  L /= 1.0*N*N;
  A /= 1.0*N*N;
  //P /= 1.0*N*N;
  //MsgBox("VSS","E:%Lf\nL:%Lf\nA:%Lf\nP:%Lf", E,L,A,P);
  delete ES;
  delete LS;
  delete AS;
  delete Block;
}


void Kernel::VSSprojection1(BYTE  * Block0, WORD N, double &E, double &P)
{
  const double Normw1 = sqrt(double(12.0));
  const double Normw2 = sqrt(double(12.0));
  const double Normw3 = sqrt(double(72.0));

  double z1,z2,z3,z4,z5,z6,z7,z8,z9, Emax=0,Eavg=0,Pmax=0,Pavg=0;
  double *Block=new double[N*N]; // block itself
  WORD I; for(I=0; I<N*N; I++) Block[I]=double(Block0[I])/255;

  E=P=0;
  for(WORD i=1; I=N*i,i<N-1; i++)
    for(WORD j=1; j<N-1; j++)
      {
      double Normz = sqrt(
              sqr(z1=Block[I-N+j-1])+sqr(z2=Block[I-N+j])+sqr(z3=Block[I-N+j+1])+
              sqr(z4=Block[I+j-1])+sqr(z5=Block[I+j])+sqr(z6=Block[I+j+1])+
              sqr(z7=Block[I+N+j-1])+sqr(z8=Block[I+N+j])+sqr(z9=Block[I+N+j+1])
              );
      if (Normz<1e-20) { continue; } // we see 0 as belonging to average subspace, although it belongs to all.
      z1 /= Normz;
      z2 /= Normz;
      z3 /= Normz;
      z4 /= Normz;
      z5 /= Normz;
      z6 /= Normz;
      z7 /= Normz;
      z8 /= Normz;
      z9 /= Normz;

      // u = P(z,{w1,w2,w3}) = u1w1+u2w2+u3w3  |w1|=|w2|=|w3|=1
      // <w1,w2>=<w1,w3>=<w2,w3>=0 : {w1,w2,w3} orthonormal basis
      double u1 =(z1+2*z2+z3 -z7-2*z8-z9)/Normw1;             // <z,w1>
      double u2 =(z1-z3 +2*z4-2*z6 +z7-z9)/Normw2;            // <z,w2>
      double u3 =(-z1-z2-z3-z4+8*z5-z6-z7-z8-z9)/Normw3;      // <z,w3>
      /*
      double Normu = sqrt(sqr(u1)+sqr(u2)+sqr(u3));
      if (Normu<1e-20) { continue; }
      // u = u/|u|
      u1 /= Normu;
      u2 /= Normu;
      u3 /= Normu;
      Normu = 1.0;
      */

      double e = sqrt(sqr(u1)+sqr(u2)); // e = |P(u,{w1,w2})|
      double p = abs(u3);               // p = |P(u,{w3})|

      if (e>Emax) Emax = e;
      Eavg += e;
      if (p>Pmax) Pmax = p;
      Pavg += p;
      }
  Eavg /= 1.0*N*N;
  Pavg /= 1.0*N*N;
  E = 0.4*Eavg+0.6*Emax;
  P = 0.4*Pavg+0.6*Pmax;
  delete Block;
}

int Kernel::quant(double x, double xmin, double xmax, int n) // quantize x in n bits
{
  return int((x/(xmax-xmin))*((1<<n)-1)); // round
}

