// file: kernel.h
// needs: dib.*, image.*, lib1.h, msgbox.h, krnldef.h, mathkrnl.*, kernel.cpp
// DIP kernel
// IFS module
// Author: Giovambattista Pulcini
// Written: 1994.12.10
// Revised:
//      5.10 - Oct 1995 32bit Flat memory model & NT port
//      5.01 - Mar 1995

//Notes:
/*
    I stripped out all "huge" pointers, so this will not run anymore on
    Windows 3.xx unless the WIN32S extensions. Besides, the code is now
    much faster, readable and somewhat portable. (May be to  OS/2, W95, 
    some UNIXes).

    Since the 5.10 Rel. is a reduced version of the full IFSAF, a lot of 
    virtual functions are declared but never implemented (well, ported)
    and used. The full porting is in progress.

    May be this would be portable under DOS, running this code along with 
    a DOS EXTENDER such as Rational DOS4GW;

    Some assumption made here may seem restrictive on processors other than
    x86 family. So we had to align structures in huge areas to segment
    boundary: from there the separation of Block-Bits from related Block Info.
    * This is no more true. But the separation is still here.

    On all the package, some structures require to be "packed", so you should
    not select align optimizations. Conditional defines or pragmas will be
    introduced in the future for supporting alignment.

*/

#ifndef __KERNEL_H__
#define __KERNEL_H__

#ifndef __MSGBOX_H__
#include "msgbox.h" // also includes STDIO, WINDOWS
#endif

#include "timecur.h"

#ifndef __IMAGE_H__
#include "image.h"
#endif

#ifndef __LIB1_H__
#include "lib1.h"
#endif

// This should be included last, so that other .h can define
// macro declared inside. It is used only for DOS environment
#ifndef __KRNLDEF_H__
#include "krnldef.h"
#endif

#include "mathkrnl.h"

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <dos.h>


#include "packing.h"


//  Class definition for gray-scale images with 256 levels 8bit/pixel

extern char far * ClassSchemeNames[15];
extern char far * BlockTypeNames[15];
extern char far * BlockTypeShorts[4];
extern char far * KrnlMetricsNames[10];
extern char far * KrnlApproxNames[10];

enum TBlockType { SHADE, MIDRANGE, EDGE, MULTIEDGE }; //int
enum TClassScheme{ CS_HISTOGRAM, CS_GRADIENT, CS_VSS, CS_TWEENPEAKS, CS_LAST=CS_TWEENPEAKS }; //int
enum TMetrics { MS, RMS, SNR, SNRMS, SNRRMS, SNRMSdB, SNRRMSdB, ABSM, MAX, MAXM, M_LAST=MAXM }; // int
enum TAppType { APPT_DEG0, APPT_DEG1, APPT_DEG2, APPT_DEG3, APPT_BILIN, APPT_BIQUAD, APPT_LAST=APPT_BIQUAD };
enum TIsometries { iso_IDENTITY, iso_FLIPV, iso_FLIPH, iso_1stDIAG, iso_2ndDIAG,
                   iso_ROT90, iso_ROT180, iso_ROT270, iso_LAST=iso_ROT270 };

enum ecErrors {
  ec_noerror  = 0x00000000,
  ec_noreport = 0x00010000,
  ec_uvdims   = 0x00020000,
  ec_nomemory = 0x00030000 
};

struct TDpool // size is 32 bytes
  {
    BYTE MeanGray;
    BYTE BlockType;
    BYTE id,jd;
    double WSum;   //8
    double WSum2;  //8
    BYTE SubMean[4]; // used for orientation scheme
    BYTE SubType[4];
    BYTE Orientation; // used for orientation scheme
    BYTE Fillup[3]; // is actually unused: we need it to align data.
  };
struct TDpoolR // size is 32 bytes
  {
    BYTE BlockEnable : 1;
    BYTE BlockType   : 2;
    BYTE Orientation : 5;     //1
    BYTE id,jd;               //2
    struct SubBlockStruct     //7*4
      { WORD  Sum;
        DWORD Sum2;
        BYTE Enable : 1;
        BYTE Type   : 2;
        BYTE Orientation : 5;
      } SubBlock[4];
    BYTE Fillup; // is actually unused: we need it to align data.
  };
struct TRpool // size is 32 bytes
  {
    BYTE MeanGray;
    BYTE BlockType;
    double WSum;
    double WSum2;
    BYTE SubMean[4];
    BYTE Orientation;
    BYTE Fillup[9]; // is actually unused: we need it to align data.
  };
struct TRpoolR // size is 32 bytes
  {
    BYTE BlockEnable : 1;
    BYTE BlockType   : 2;
    BYTE Orientation : 5;     //1
    struct SubBlockStruct     //7*4
      { WORD  Sum;
        DWORD Sum2;
        BYTE Enable : 1;
        BYTE Type   : 2;
        BYTE Orientation : 5;
      } SubBlock[4];
    BYTE Fillup[3]; // is actually unused: we need it to align data.
  };

struct TAppRes // Approximation results
  {
    double dblError;          // Global error (sum) over all image
    double dblThreshold;      // ErrorThreshold value @ calc time
    WORD wCount;              // Number of block currently approximated
  };

_CLASSDEF(TEncStats)
struct TEncStats // Encoding Image...
  {
    DWORD dwIFSBlk, dwPolyBlk, dwSubBlk, dwCountBlk,
          dwAlphagt1,   dwAppAlphagt1,
          dwDet0,       dwAppDet0,
          dwAlphalt0,   dwAppAlphalt0,
          dwAlphaltm1,  dwAppAlphaltm1,
          dwAlphaCount, dwAppAlpha,
          dwSearch,
          dwLSSsuccess, dwLSSsameiso;
    DWORD isofrequency[iso_LAST+1];
    DWORD packed, original;
    DWORD ElapsedTime;
    double ErrorThreshold;
    double GlobalError, GlobalErrorwoLSS;
    char szOutFileName[80];
    char szInFileName[80];

    // Kernel settings
    WORD B,W,H,did,djd,nd,md,nr,mr,overstep;
    DWORD RIsize, DIsize, Dsize;
    WORD Metric, ClassScheme, Approx;
    double EdgeThreshold, MidrangeThreshold, ShadeThreshold;
    double w[3];
    double alphamin,alphamax,betamin,betamax;
    WORD alphabits,betabits,polybits;
    WORD ENC_USEPOLY,ENC_QUANTIZEALPHA,ENC_BOUNDALPHA,ENC_QUANTIZEPOLY,
         ENC_POLYONLY, ENC_USEISOMETRIES, ENC_STRIPSHADES, ENC_OVERSTEP,
         ENC_OVERSTEPUSEISOMETRIES, ENC_USEORIENTATIONSCHEME;

    WORD wErrRange[15];
    WORD wAlpha[100];
    WORD wBeta[100];
    WORD wDeltax[100];
    WORD wDeltay[100];
    WORD wDist[100];

    BYTE EncodeAlgorithm;
    WORD SimAnnSize;

    // Recursion support chunk
    WORD wSubBlock[4];
    WORD MaxLevel; // Max level of recursive subdivision
    int diameter;
    double ErrorThresholdR;
  };

struct TDecStats
  {
    char szFileName[80];
    WORD W,H;  // decode size
    double rmsError, deltarmsError, SNR;
  };

struct TSubBlkParm;

struct IFSnode  { double x; double y; int iso; double alpha; double beta; };
struct Polynode { int n; double * c; };
typedef IFSnode * PTIFS;
typedef Polynode * PTPoly;
struct Transfnode
{
  BOOL Polynomial;
  union { PTIFS IFS;
          PTPoly Poly;
          };
};

struct IFSsubnode { double x; double y; int iso; double alpha; double beta; BYTE k; };
typedef IFSsubnode far * PTIFSSR;
struct IFSnodeR  { double x; double y; int iso; double alpha; double beta; WORD sbb; PTIFSSR sub[4]; };
typedef IFSnodeR far * PTIFSR;
struct TransfnodeR
{
  BOOL Polynomial;
  union { PTIFSR IFS;
          PTPoly Poly;
          };
};
struct TSubBlkParmR
  {
    int idbits,jdbits;
    int W,H,B;
    WORD alphabits, betabits,did,djd;
    double alphamax, betamax, alphamin, betamin;
    BOOL DEC_QUANTIZEALPHA;
    double ScaleFactorx,  ScaleFactory;
  };

// typedef BYTE   * LPBYTE;  // already defined in image.h
typedef BYTE   * HPBYTE; // huge pointer to byte
struct cDPoolInfo {
  DWORD SSum[4];
  DWORD Sum2;
  WORD  Sum;
  BYTE SOrientation[4];
  BYTE Orientation;
  BYTE fillup[3];
  WORD fillup1;
};
typedef cDPoolInfo   * HPDPI;


typedef double TMatrix[64]; // max 32x32
typedef double far * LPDBL;
  
enum QuantizeClass { ALPHAY, ALPHAI, ALPHAQ, BETAY, BETAI, BETAQ, QCLASSES};


_CLASSDEF(Kernel)
class far Kernel
{
public:
  DWORD Timing; // elapsed time (in ticks of 1 ms) of last non trial operation
  DWORD ERRORCODE;
  BOOL Verbose; // be-verbose flag
  BOOL Changes; // true if some of below params are changed.
  WORD B;  // Range Block size
  WORD D;  // Domain Block size
  WORD MaxLevel; // Max level of recursive subdivision
  WORD did,djd,overstep;    // step
  int diameter, radius;
  WORD alphabits, betabits, polybits;
  WORD SimAnnSize;
  WORD decsteps;
  TClassScheme ClassScheme;  // Blocks classifing shceme
  TAppType ApproximationType;
  TMetrics Metric;
  double EdgeThreshold, MidrangeThreshold, ShadeThreshold; // x<yThreshold is y
  LPDBL Basis, Weight, invG;
  double ErrorThreshold, ErrorThresholdR;
  BOOL ENC_USEPOLY, ENC_POLYONLY, ENC_USEISOMETRIES, ENC_BOUNDALPHA,
       ENC_QUANTIZEALPHA, ENC_STRIPSHADES, ENC_QUANTIZEPOLY,
       ENC_OVERSTEP, ENC_OVERSTEPUSEISOMETRIES, ENC_USEORIENTATIONSCHEME;
  BOOL DEC_SINGLESTEP, DEC_SAVESTEPS;
  double alphamin, alphamax, betamin, betamax;

  double quantizemin[QCLASSES],quantizemax[QCLASSES],
  	 quantizestep[QCLASSES],quantizeinvstep[QCLASSES];
  WORD quantizebits[QCLASSES];

  BYTE IsometryClass[24][24];
  BYTE EncodeAlgorithm;
  FILE * rpt;

// x_step = 1.0/(x_max-x_min)*((1ul<<x_bits)-1)
// x_invstep= 1.0/double((1ul<<x_bits)-1)*(x_max-x_min)=1/x_step
inline DWORD Quantize(double x, int k)
  { return (DWORD(0.5 + (x-quantizemin[k])*quantizestep[k])); }
inline double Dequantize(DWORD x, int k)
  { return (double(x)*quantizeinvstep[k]+quantizemin[k]); }


  void QuantizationSetup()
	{
	  for(int k = 0; k<QCLASSES; k++)
	    {
	      quantizestep[k] = 1.0/(quantizemax[k]-quantizemin[k])*((1ul<<quantizebits[k])-1);
	      quantizeinvstep[k] = 1.0/quantizestep[k];
	    }
	}

  void Reset()
    { ERRORCODE = ec_noerror; 
      B=8; D=B*2; MaxLevel=2;
      did=1; //B/4;
      djd=1; //B/4;
      //nr=mr=nd=md=0;
      polybits=10;
      ClassScheme=CS_HISTOGRAM;
      Verbose = TRUE;
      Changes = FALSE;
      EdgeThreshold = 1.0;
      MidrangeThreshold = 2.0/4;
      ShadeThreshold = 1.0/4;
      ApproximationType = APPT_DEG0;
      Metric = MS;
      Timing = 0;
      ErrorThreshold  = 10.0;
      ErrorThresholdR = 4.0;  // in rms value
      SimAnnSize = 10;
      decsteps = 10;
      ENC_USEPOLY = TRUE;
      ENC_POLYONLY= FALSE;
      ENC_BOUNDALPHA=TRUE;
      ENC_USEISOMETRIES=TRUE;
      ENC_QUANTIZEALPHA=TRUE;
      ENC_QUANTIZEPOLY=TRUE;
      ENC_STRIPSHADES=TRUE;
      ENC_OVERSTEP = FALSE;
      ENC_OVERSTEPUSEISOMETRIES=FALSE;
      ENC_USEORIENTATIONSCHEME=FALSE;
      overstep = 1;
      diameter = 5;
      radius = 15;
      DEC_SINGLESTEP=FALSE;
      DEC_SAVESTEPS=FALSE;
      alphamin = -1;
      alphamax = 1;
      alphabits = 8;
      betabits  = 9;
      betamin = -255.0;
      betamax =  255.0;
      //idbits = 8;
      //jdbits = 8;
      EncodeAlgorithm = 0; // 0:Full search, 1:SA, 2:GENETIC
      
      // qunatization settings & inits
	  quantizemin[ALPHAY] = -1.0;    quantizemax[ALPHAY] = 1.0;
	  quantizemin[ALPHAI] = -1.0;    quantizemax[ALPHAI] = 1.0;
	  quantizemin[ALPHAQ] = -1.0;    quantizemax[ALPHAQ] = 1.0;
	  quantizemin[BETAY] = -255.0;   quantizemax[BETAY]  = 255.0;
	  quantizemin[BETAI] = -355.0;   quantizemax[BETAI]  = 355.0;
	  quantizemin[BETAQ] = -355.0;   quantizemax[BETAQ]  = 355.0;

	  quantizebits[ALPHAY] = 8;
	  quantizebits[ALPHAI] = 8;
	  quantizebits[ALPHAQ] = 8;
	  quantizebits[BETAY ] = 8;
	  quantizebits[BETAI ] = 8;
	  quantizebits[BETAQ ] = 8;

       QuantizationSetup();

    };


  void (*Processfn)(char *, float);
  
  Kernel()
    { Reset();
      Basis = Weight = invG = NULL;
      Processfn = NULL;
    };
  ~Kernel()
    {
      if (Basis) delete Basis;
      if (Weight) delete Weight;
      if (invG) delete invG;
    };

  BOOL ReadMathStream(char * filename, WORD deg, WORD coeff, WORD N /* , double * & Basis, double * & Weight, double * & invG */ );



  // Block extraction
  void GetImageBlock(const RTImage img, WORD i, WORD j, WORD N, BYTE * Block) {img.GetBlock(i,j,N,Block);}; //newer
  void ExtractBlock(const RTImage img, WORD i0, WORD j0, WORD N, BYTE   * Bits); // Extract a block NxN from image
  void ExtractChild(PTBLOCK Block, WORD N, PTBLOCK Child[4]);


  // Block classification
  void ImageBlockClassify(const RTImage img, DWORD Blocks[4]);// Classify all blocks, makes an histogram
  BYTE BlockHistogram(const RTImage, WORD ib, WORD jb, DWORD * Histogram); // Calculate histogram of B-block, returns MEAN value of block
  TBlockType BlockClassify(const RTImage img, WORD ib, WORD jb); // Classify B-Block(ib,jb) with ib in {0..W/B}, jb in {0..H/B}
  TBlockType Classify(BYTE   * Bits, WORD N); // Classify N-block pointed by Bits
  void BlockSum(const RTImage img, WORD ib0, WORD jb0, double & WSum, double & WSum2, double SSum[4]); // returns sum and sum^2 of block

  // Block processing
  void Contract(const RTImage img, WORD id, WORD jd, BYTE   * DPool, BYTE   * DPoolInfo); // Make a D-Pool block from image
  double Approximate(PTBLOCK A, PTBLOCK B, WORD N, double * p);
  void AbsDifference(PTBLOCK A, PTBLOCK B, PTBLOCK E, WORD N); // computes E = |A-B|
  void Difference(PTBLOCK A, PTBLOCK B, PTBLOCK E, WORD N); // computes E = A-B

  void MakeApproxBlock(WORD m, LPDBL c, WORD N, PTBLOCK B); // computes an approximation of B
  void BlockLumScale(PTBLOCK A, PTBLOCK B, WORD N, double alpha, double beta);
  void BlockTransform(PTBLOCK A, PTBLOCK B, WORD N, TIsometries isometry);
  void ChildBlockEncode(WORD ir, WORD jr, WORD N, PTBLOCK Block);

  void MakeApproxBlock(WORD m, LPDBL c, WORD N, LPDBL B);
  void ConvertBlock(double   * A, BYTE   * B, WORD N);
  //void SetBlockToImage(double   * Image, WORD W, WORD H, WORD i0, WORD j0, double   * Block, WORD N);
  //void GetBlockFromImage(double   * Image, WORD W, WORD H, WORD i0, WORD j0, double   * Block, WORD N);
  void SetDibToImage(const RTImage img, double   * Image, WORD W, WORD H);
  void Contract(double   * A, double   * B, WORD N);
  void BlockTransform(double   * A, double   * B, WORD N, TIsometries isometry);
  void BlockLumScale(double   * A, double   * B, WORD N, double alpha, double beta);
  double BlockError(PTBLOCK A, PTBLOCK B, WORD N,double alpha, double beta);
  double MSDifference(double * A, double * B, WORD N);
  double MSDifference(const RTImage img, WORD i0, WORD j0, BYTE * Block, WORD N);

  void ContractR(const RTImage img, WORD id0, WORD jd0, BYTE   * DPool, TDpoolR   * DP);
  void MinimizeCRms( TEncStats &S, int ka, int kb,
                          WORD N, double SumR, double SumR2, double SumD, double SumD2, double SumRD,
                          DWORD &ialpha, DWORD &ibeta, double &alpha, double &beta, double&E );
  void MakeDomainPoolB1(const RTImage img, BYTE   * DPool, TDpool   * DPoolInfo, DWORD &DPoolSize, DWORD Blocks[4]);
  void EncodeImageB1(TEncStats &S, LPSTR szDefName, const RTImage img1, RTImage img2, TRpool   * RPoolInfo, TDpool   * DPoolInfo, PTBLOCK DPool);
  void EncodeImageR1(TEncStats &S, LPSTR szDefName, const RTImage img1, RTImage img2, TRpoolR   * RPoolInfo, TDpoolR   * DPoolInfo, PTBLOCK DPool);
  void EncodeImageCR1(TEncStats &S, LPSTR szDefName, const RTImage img1, RTImage img2, TRpoolR   * RPI, TDpoolR   * DPI, PTBLOCK DPool);
  void InitEncStats( TEncStats &S, char * szInFileName, char * szOutFileName,
                     WORD W, WORD H, WORD nr, WORD mr, WORD nd, WORD md, WORD idbits, WORD jdbits,
                     DWORD ID, DWORD RPIS, DWORD DPIS, DWORD DPS);
  void WriteReportHeader( FILE * rpt, TEncStats & S, WORD idbits, WORD jdbits, DWORD ID);
  void UpdateEncStats( TEncStats &S, TIsometries isoBest,
                             double alphaBest, double betaBest, double E,
                             WORD ir, WORD jr, WORD id, WORD jd);
  void FixupEncStats(TEncStats & S, double GlobalError, double GlobalErrorwoLSS);
  void WriteReportFooter(TEncStats & S, double GlobalError, double GlobalErrorwoLSS);
  BOOL DecodeImageCR1(TDecStats * DS, LPSTR szDefName, PTImage img0, const RTImage img1, RTImage img2, WORD Passes, int FileNameCount, BOOL FirstDecode);



  // Block metrics
  double Distance(BYTE   * A, BYTE   * B, WORD N); // compute distance between two blocks
  double BlockCrossSum(PTBLOCK A, PTBLOCK B, WORD N); // returns sum of w*A*B
  int quant(double x, double xmin, double xmax, int n); // quantize x with n bits
  int RangeCheck(double); // assure return value beeing in range 0..255

  // Block Image processing
  double Gradx(BYTE   * sBlock, BYTE   * dBlock, WORD N);
  double Grady(BYTE   * sBlock, BYTE   * dBlock, WORD N);
  void Grad(BYTE   * Gx, BYTE   * Gy, BYTE   * dBlock, WORD N);
  void Laplacian(BYTE   * sBlock, BYTE   * dBlock, WORD N);
    // Projection on Edge, Line, Average, isolated Pixel vector sub-spaces
  void VSSprojection(BYTE   * Block, WORD N, double &E, double &L, double &A, double &P);
  void VSSprojection1(BYTE   * Block0, WORD N, double &E, double &P); // My implementation

  //Metrics
  BOOL CompatibleImage(const RTImage img) // verifies if all conditions match current image
    {return (img.wWidth%B==0 && img.wHeight%B==0); //check gray-scale also, minimal dimensions
    };
  void GetDomainBlocksNumber(const RTImage img, WORD & nd, WORD & md)
    { nd = (img.wHeight-D)/did+1; md = (img.wWidth-D)/djd+1; };
  void GetRangeBlocksNumber(const RTImage img, WORD & nr, WORD & mr)
    { nr = img.wHeight/B; mr = img.wWidth/B; };

  // D,R pool
  // Before calling MakeDomainPool, allocate a D-pool array of size
  // GetDPoolSize, and an array of info of GetDPoolInfoSize
  DWORD GetDPoolSize(const RTImage img); // returns bytes needed to store D-Pool
  DWORD GetDPoolInfoSize(const RTImage img); // returns bytes needed to store D-Pool info
  DWORD GetRPoolInfoSize(const RTImage img); // returns bytes needed to store R-Pool info
  void MakeDomainPool(const RTImage img, BYTE   * DPool, TDpool   * DPoolInfo, DWORD &DPoolSize, DWORD Blocks[4]);
  void MakeRangePool(const RTImage img, TRpool   * RPoolInfo, DWORD Blocks[4]);
  TAppRes ApproximateImage(const RTImage img1, RTImage img2); // makes block approximation, return total error


  void WriteHeader(FILE * cod, DWORD ID, WORD W, WORD H, WORD B, WORD did, WORD djd,
                             WORD idbits, WORD jdbits, WORD ENC_QUANTIZEALPHA,
                             WORD alphabits, WORD betabits, WORD polybits, WORD ENC_QUANTIZEPOLY);
  void ReadHeader(FILE * cod, DWORD &ID, WORD &W, WORD &H, WORD &B, WORD &did, WORD &djd,
                             WORD &idbits, WORD &jdbits, WORD &DEC_QUANTIZEALPHA,
                             WORD &alphabits, WORD &betabits, WORD &polybits, WORD &DEC_QUANTIZEPOLY);
  void WritePolyCoeff(FILE * cod, double * p, int n);
  void ReadPolyCoeff(FILE * cod, double * p, int n, BOOL DEC_QUANTIZEPOLY);
  void WriteQuantizedPoly(FILE * cod, double * p, int n);
  void EncodeImage(TEncStats &, LPSTR szDefName, const RTImage img1, RTImage img2, TRpool   * RPoolInfo, TDpool   * DPoolInfo, PTBLOCK DPool);
  void EncodeImageSA(TEncStats &, LPSTR szDefName, const RTImage img1, RTImage img2, TRpool   * RPoolInfo, TDpool   * DPoolInfo, PTBLOCK DPool);
  void EncodeImageSA4(TEncStats &, LPSTR szDefName, const RTImage img1, RTImage img2, TRpool   * RPoolInfo, TDpool   * DPoolInfo, PTBLOCK DPool);
  void EncodeImageR3SA(TEncStats &S, LPSTR szDefName, const RTImage img1, RTImage img2, TRpoolR   * RPI, TDpoolR   * DPI, PTBLOCK DPool);
  BOOL DecodeImage(TDecStats *, LPSTR szDefName, PTImage img0, const RTImage img1, RTImage img2, WORD Passes, int FileNameCount, BOOL FirstDecode);
  void DecodeImage3(const RTImage img1, RTImage img2, WORD Passes, int FileNameCount=0);
  BOOL DecodeImage4(TDecStats *, LPSTR szDefName, PTImage img0, const RTImage img1, RTImage img2, WORD Passes, int FileNameCount, BOOL FirstDecode);
  BOOL DecodeImageR(TDecStats * DS, LPSTR szDefName, PTImage img0, const RTImage img1, RTImage img2, WORD Passes, int FileNameCount, BOOL FirstDecode);
  void ReadCodeR(LPSTR szFileName, TransfnodeR   * T, WORD &K);
  void SubBlockDecode(TDecStats * DS, TSubBlkParm& sbp, FILE * cod, int K, double * RBlock, double * Block, double * DBlock, double   * image );
  BOOL DecodeImage4R(TDecStats * DS, LPSTR szDefName, PTImage img0, const RTImage img1, RTImage img2, WORD Passes, int FileNameCount, BOOL FirstDecode);
  void SubBlockDecodeR(TSubBlkParmR& sbp, PTIFSSR T, int K, double * RBlock, double * Block, double * DBlock, double   * Image );
  void ReadCode(LPSTR szFileName, Transfnode   * T, WORD &K, FILE * rpt);
  void MakeApproxBlock1(WORD m, LPDBL c, WORD M, WORD N, LPDBL B);
  void MinimizeRms(TEncStats & S, PTBLOCK RBlock, PTBLOCK DBlock, WORD N, double SumR, double SumR2, double SumD, double SumD2,
                                 double &alpha, double &beta, DWORD &ialpha, DWORD &ibeta, double &E, double&iE );
  void MakeDomainPoolR(const RTImage img, BYTE   * DPool, TDpoolR   * DPI, DWORD &DPoolSize, DWORD Blocks[4]);
  void BlockSumR(const RTImage img, WORD ib0, WORD jb0, WORD Sum[4], DWORD Sum2[4], WORD SSum[4][4]);
  void MakeRangePoolR(const RTImage img, TRpoolR   * RPI, DWORD Blocks[4]);
  void SubBlockEncode( TEncStats &S, FILE * cod, int K, PTBLOCK RBlock,
                             WORD ir, WORD jr, WORD idRBest, WORD jdRBest,
                             TRpoolR far * RPI, TDpoolR   * DPI, PTBLOCK DPool,
                             RTImage img, double E[4]);
  void SubBlockEncodeSA( TEncStats &S, FILE * cod, int K, PTBLOCK RBlock,
                             WORD ir, WORD jr, WORD idRBest, WORD jdRBest,
                             TRpoolR far * RPI, TDpoolR   * DPI, PTBLOCK DPool,
                             RTImage img, double E[4]);
  void SubBlockEncodeC( TEncStats &S, FILE * cod, int K, PTBLOCK RBlock,
                             WORD ir, WORD jr, WORD idRBest, WORD jdRBest,
                             TRpoolR far * RPI, TDpoolR   * DPI, PTBLOCK DPool,
                             RTImage img, double SE[4]);
  void DoPrecalculation(BYTE   * lpDibRGB, HPBYTE &lpY, HPBYTE &lpI, HPBYTE &lpQ,
                                              HPBYTE &lpcY,HPBYTE &lpcI,HPBYTE &lpcQ,
                                              HPDPI &DPI);





  //GLOBAL IMAGE TRANSFORMS
  // Full Image processing
  double sqrSum(const RTImage img);
  void GrayLevelScale(RTImage img, int n, int d, int L);
  DWORD MakeHistogram(const RTImage img, DWORD * Histogram);  // Calculate histogram of image, returns its max value

  // some adds
  void CalcScheme();

  char * GetTimeString();

  // Interactive function, front-end specific
  // they cannot be static if virtual, but we
  // need not they to be virtual.
  HCURSOR hcSavedCursor;

  virtual void ShowBegin() { 
       hcSavedCursor = SetCursor(LoadCursor(0, IDC_WAIT));  
       SetClockCursor(0); };
  virtual void ShowRatio(float x) 
    { SetClockCursor(int(x*100)); 
      if (Processfn) Processfn(NULL, x);
    }; 
  // virtual void ShowMessage(char *, ...) {}; // NOT ALLOWED BY WATCOM
  virtual void ShowMessage(char * msg) 
    { if (Processfn) Processfn(msg, -1); };
  virtual void ShowEnd() { SetCursor(hcSavedCursor); };

#pragma off (unreferenced)
  virtual void InitImage(TImage *, BOOL info = TRUE) {};
  virtual void UpdateImage(double r, WORD ir, WORD jr, double *p, int n, double E, double GE) {};
  virtual void UpdateImage(double r, WORD ir, WORD jr, WORD id, WORD jd, int iso, double a, double b, double E, double GE) {};
  virtual void UpdateImage(double r = 0) {};
  virtual void EndImage() {};
  virtual void Beep() {};
  virtual DWORD GetTimeTicks() { return GetTickCount(); }; // return real time ms counter
  virtual void * GetHugeMem(DWORD size) { return (void *) malloc(size); };
  virtual void FreeHugeMem(void * & P) { free(P); P=NULL; };
#pragma on (unreferenced)
};

  //Image Inits & Block Extraction
  //BOOL ReadMatrix(char * filename, WORD N, TMatrix & matrix);
  //void WriteMatrix(char * filename, WORD N, TMatrix matrix);
  BOOL ReadVector(FILE * in, WORD N, LPDBL v);
  BOOL ReadMatrix(FILE * in, WORD M, WORD N, LPDBL m);
  void WriteVector(FILE * out, WORD N, LPDBL v);
  void WriteMatrix(FILE * out, WORD M, WORD N, LPDBL m);
  void MultiplyMatrixVector(WORD M, WORD N, LPDBL m, LPDBL v1, LPDBL v2);
  double DotVector(WORD N, LPDBL x, LPDBL y, LPDBL w = NULL);
  void NormalizeWeightVector(WORD N, LPDBL w); // Normalize a weight vector

  void SetBlockToImage(double   * Image, WORD W, WORD H, WORD i0, WORD j0, double   * Block, WORD N);
  void GetBlockFromImage(double   * Image, WORD W, WORD H, WORD i0, WORD j0, double   * Block, WORD N);
  BYTE GetOrientation(WORD SSum[4]);
  void BlockSubError(PTBLOCK A, PTBLOCK B, WORD N, double E[4]);
  void GetSubBlock(int k, PTBLOCK A, WORD M, PTBLOCK B, WORD N);

#endif
