// lib1.h
// Library functions
// Written by G.Pulcini
// Last revised 1995/2/13  Rev. 1.01

#ifndef __LIB1_H__
#define __LIB1_H__

inline int Fatt(int n)
  { if (n==0 || n==1) return 1;
    int s=1; for(int k=2; k<=n; k++) s*=k;
    return s;
  }

inline double sqr(double a) {return a*a; }

template <class T> Swap(T & a, T & b) { T temp=a; a=b, b=temp; };
template <class T> T Bound(T x, T lx, T hx) { if (x<lx) x=lx; if (x>hx) x=hx; return x; };
template <class T> T Min(T x, T y) { return (x<y ? x:y); };
template <class T> T Max(T x, T y) { return (x>y ? x:y); };
template <class T> int Permutation(T a[], int n)
// returns lexicographic order of permutation
{
  int k,swaps,s=0;
  for(k=0; k<n-1; k++)
    {
    swaps=0;
    for(int i=k+1; i<n; i++) if (a[k]>a[i]) swaps++;
    s += swaps*Fatt(n-k-1);
    }
  return s;
};
template <class T> BOOL Permute( T * a, int n )
// Returns FALSE if the last permutation in lexicographic order has been generated.
// You should start with permutation a[i]=i, test it and then call Permute(a,n)
// testing if it is the last (FALSE returned) or not.
// Here is an example:
//    a[i]=i; TEST(a,n); while(Permute(a,n)) TEST(a);
{
  if (n<2) return FALSE;  // no permutations on less than 2 elements
  // for(BOOL stop=FALSE; !stop; )
  BOOL stop=FALSE;
  {
    // find the biggest i such that a[i]<a[i+1]
    int i=n-2;                                          // Here we assume n>=2
    while ((i>=0) && (a[i]>=a[i+1])) i--;
    // here i=-1 means that there are no other possibilities.
    stop = (i==-1); 
    if (stop) return (!stop); // since below things may go wrong with i==-1

    // find a[j] such that a[j] is the littlest element
    // greater than a[i] between the ones at the right of i
    int j=i+1; // actually is a[i]<a[j=i+1]
    for (int k=j; k<n; k++) if ((a[k]<a[j]) && (a[k]>a[i])) j=k;

    // swap a[i] and a[j]
    Swap(a[i],a[j]);     // int t=a[i]; a[i]=a[j]; a[j]=t;

    // int h = (n-i)/2; number of elements to swap
    // for k 1..h do swap a[i+k] with a[n-k+1]
    for (k = 1; k<=(n-(i+1))/2; k++) Swap(a[i+k],a[n-k]);
      //{ t=a[i+k]; a[i+k]=a[n-k]; a[n-k]=t; } // ((i+1)+k)-1 = i+k,  (n-k+1)-1 = n-k
  }
  return !stop;
};

#endif
