// file: crdecode.cpp
// color Recursive decoding, 2 levels
// Written by G.Pulcini
// Last revised 1995/4/20  Rev 0.30

#include "rcencdec.h"

#define GlobalAllocPtr(x,y) malloc(y)
#define GlobalFreePtr(x) free(x)

// void SubBlockDecode(TDecStats * DS, FILE * cod, int K, double * Block);

struct TSubBlkParm
  {
    int idbits,jdbits;
    int W,H;
    BOOL DEC_QUANTIZEALPHA;
  };

typedef double  * HPDOUBLE;


void ReadColorHeader(FILE * cod, DWORD &ID, WORD &W, WORD &H, WORD &B, 
			WORD &did, WORD &djd, int & dRadius,
			double * quantizemin, double * quantizemax,
			WORD * quantizebits)
{
  Unpack(-2, cod);  // init
  ID = (DWORD) Unpack(32, cod);
  W  = (WORD)  Unpack(16, cod);
  H  = (WORD)  Unpack(16, cod);
  B  = (WORD)  Unpack(8, cod);
  did= (WORD)  Unpack(8, cod);
  djd= (WORD)  Unpack(8, cod);
  dRadius=(WORD) Unpack(16, cod);
  for(int k=0; k<QCLASSES; k++)
    {
      quantizemin[k] = Unpackdouble(cod);
      quantizemax[k] = Unpackdouble(cod);
      quantizebits[k]= (BYTE) Unpack(8, cod);
    }
}



void GetBlock( HPDOUBLE lpcY, WORD wbits, WORD N, int iso,
                      double alpha, double beta,
                      WORD id0, WORD jd0, double * Block)
{
  register WORD i,j,I;
  DWORD Id;
  switch(iso)
  {
  case iso_IDENTITY:
     for(i=0; I=i*N, Id=(DWORD(id0+i)<<wbits)+jd0, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = lpcY[Id+j]*alpha+beta;
     break;
  case iso_FLIPV:
     for(i=0; I=i*N, Id=(DWORD(id0+i)<<wbits)+jd0, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = lpcY[Id+N-1-j]*alpha+beta;
     break;
  case iso_FLIPH:
     for(i=0; I=i*N, Id=(DWORD(id0+(N-1-i))<<wbits)+jd0, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = lpcY[Id+j]*alpha+beta;
     break;
  case iso_1stDIAG:
     for(i=0; I=i*N, i<N; i++)
       for(j=0; Id=(DWORD(id0+j)<<wbits)+jd0, j<N; j++)
         Block[I+j] = lpcY[Id+i]*alpha+beta;
     break;
  case iso_2ndDIAG:
     for(i=0; I=i*N, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = lpcY[(DWORD(id0+(N-1-j))<<wbits)+jd0+(N-1-i)]*alpha+beta;
     break;
  case iso_ROT90:
     for(i=0; I=i*N, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = lpcY[(DWORD(id0+j)<<wbits)+jd0+(N-1-i)]*alpha+beta;
     break;
  case iso_ROT180:
     for(i=0; I=i*N, Id=(DWORD(id0+(N-1-i))<<wbits)+jd0, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = lpcY[Id+(N-1-j)]*alpha+beta;
     break;
  case iso_ROT270:
     for(i=0; I=i*N, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = lpcY[(DWORD(id0+(N-1-j))<<wbits)+jd0+i]*alpha+beta;
     break;
  }
}

void SetBlock(HPDOUBLE lpY, WORD Wbits, WORD N, WORD i0, WORD j0, double * Block)
{
  register WORD i,j,I;
  DWORD Id;
  for(i=0; I=i*N, Id=(DWORD(i0+i)<<Wbits)+j0, i<N; i++)
    for(j=0; j<N; j++)
      lpY[Id+j] = Block[I+j];
}

void CombineYIQBlock2RGB(double * BlockY, double * BlockI, double * BlockQ, PTBLOCK BlockRGB, WORD N)
{
  WORD I1,I2;
  for(WORD i=0; I2=i*N*3, I1=i*N, i<N; i++)
    for(WORD j=0; j<N; j++)
      {
        int R,G,B;
        WORD P = I2+(j<<1)+j, P1 = I1+j;
        YIQtoRGB(BlockY[P1], BlockI[P1], BlockQ[P1], R, G, B);
        BlockRGB[P]   = Bound(R, 0, 255);
        BlockRGB[P+1] = Bound(G, 0, 255);
        BlockRGB[P+2] = Bound(B, 0, 255);
      }
}

void ContractImage(HPDOUBLE lpY, HPDOUBLE lpI, HPDOUBLE lpQ, WORD Wbits, HPDOUBLE lpcY, HPDOUBLE lpcI, HPDOUBLE lpcQ, WORD wbits, WORD w, WORD h)
{
  DWORD dwSI,dwSI1,dwDI,dwDP,P0,P1,P2,P3,J;
  for(WORD i=0; dwSI=DWORD(i<<1)<<Wbits, dwSI1=DWORD((i<<1)+1)<<Wbits, dwDI=DWORD(i)<<wbits, i<h; i++)
  {
    //ShowRatio(float(i)/h);
    for(WORD j=0; j<w; j++)
      {
        J = j<<1;
        P0 = dwSI|J;
        P1 = dwSI|(J+1);
        P2 = dwSI1|J;
        P3 = dwSI1|(J+1);
        dwDP = dwDI|DWORD(j);

        lpcY[dwDP] = (lpY[P0]+lpY[P1]+lpY[P2]+lpY[P3])*0.25;
        lpcI[dwDP] = (lpI[P0]+lpI[P1]+lpI[P2]+lpI[P3])*0.25;
        lpcQ[dwDP] = (lpQ[P0]+lpQ[P1]+lpQ[P2]+lpQ[P3])*0.25;
      }
  }
}

#define PROFILING

#ifdef PROFILING
 #define PROFILE(x) \
  Time1 = GetTimeTicks()-Time1; \
  fprintf(rpt, x ": %lu\n", Time1); fflush(rpt); \
  Time1 = GetTimeTicks();
#elif
 #define PROFILE(x) ;
#endif

#define ShowReset() {ShowEnd(); ShowBegin();}


BOOL Kernel::DecodeImageCR1(TDecStats * DS, LPSTR szDefName, PTImage img0, const RTImage img1, RTImage img2, WORD Passes, int FileNameCount, BOOL FirstDecode)
// img0: the original uncompressed image
// img1: the starting image for reconstruction
// img2: the recostructed image
{
  DWORD Time = GetTimeTicks();
  #ifdef PROFILING
  DWORD Time1= GetTimeTicks();
  #endif

  char szFileName[80];
  if (FirstDecode)
    strcpy(szFileName, szDefName);
  else
    sprintf(szFileName, "IFS%03u", FileNameCount, 0);
  strcat(szFileName, "D.RPT");
  FILE * rpt = fopen(szFileName,"w");

  strcpy(szFileName, szDefName);
  strcat(szFileName, ".IFS");
  FILE * cod = fopen(szFileName,"rb");
  if (cod) fprintf(rpt, "Code file opened.\n"); fflush(rpt);

  int  dRadius,k;
  DWORD ID;
  WORD cW,cH;
  ReadColorHeader(cod, ID, cW, cH, B, did, djd, dRadius, 
  	quantizemin, quantizemax, quantizebits);
  if (ID != 0x0AFFF08ul) fprintf(rpt, "BAD file ID\n");
  fprintf(rpt, "ID: %lx\n", ID);
  fprintf(rpt, "Image size: %d x %d\n", cW, cH); fflush(rpt);
  fprintf(rpt, "B:%u did:%u djd:%u\n", B, did, djd);
  fprintf(rpt, "dRadius: %u\n", dRadius);
  fprintf(rpt, "Qbits: ");
  for(k=0; k<QCLASSES; k++) fprintf(rpt, "%d ", quantizebits[k]);
  fprintf(rpt, "\n");
  fflush(rpt);
  fclose(cod);

  ShowBegin();
  ShowMessage("Converting image...");
  ShowRatio(0.0);

  // by now img1 has to be of the same size of original image!
  if (!img1.lpDib) // no source image selected
    {  // we build a black image
      img1.Setup( CreateRGBDib(cW, cH, 0,0,0) );
    }
  img2.Copy(img1);

    strcpy(DS->szFileName,"");
    DS->W=img1.wWidth;
    DS->H=img1.wHeight;
    DS->SNR=0;
    DS->rmsError=0;
    DS->deltarmsError=0;

  HPBYTE lpDibRGB = img1.lpDib;       // We should here test type of the image...

  HPDOUBLE
         lpY, lpI, lpQ,               // YIQ color components
         lpcY, lpcI, lpcQ;            // YIQ contracted color components
  HPDPI  DPI;                         // color Domain Pool Info
  HPBYTE lpRGBbits = GetDibBitsAddr(lpDibRGB);

  if ( (1u<<bitsof(djd)) != djd ||
       (1u<<bitsof(did)) != did ||
       (1u<<bitsof(B))   != B ||
        B < 2*djd ||
        B < 2*did
     ) // do something of appropriate...
     { MsgBox("Error", "Unvalid parameters #01"); return FALSE; }

  if ( (1u<<bitsof(sizeof(double)))!=sizeof(double))
     { MsgBox("Error", "Unvalid double size"); return FALSE; }

  // uppercase are relative to image
  // lowercase are relative to contracted image
  int  diameter= 2*dRadius+1;
  WORD bB = bitsof(B),
       bdid = bitsof(djd),
       bdjd = bitsof(did),
       bdbl = bitsof(sizeof(double));
  WORD W = GetDibWidth(lpDibRGB),
       H = GetDibHeight(lpDibRGB),
       Wbits = bitsof(W),
       Hbits = bitsof(H),
       W2 = WORD(1) << Wbits,
       H2 = WORD(1) << Hbits,
       mr = (W>>bB) + ((W%B)?1:0),      // W/B + ((W%B)?1:0),
       nr = (H>>bB) + ((H%B)?1:0),      // H/B + ((H%B)?1:0),
       W1 = mr<<bB,                     // mr*B,
       H1 = nr<<bB,                     // nr*B,
       w = W1>>1,
       h = H1>>1,
       wbits = bitsof(w),
       hbits = bitsof(h),
       w2 = 1u << wbits,
       h2 = 1u << hbits,
       md = (w-B)>>bdjd,                // (w-B)/djd,
       nd = (h-B)>>bdid,                // (h-B)/did,
       jdbits = Min(bitsof(diameter), bitsof(md)),
       idbits = Min(bitsof(diameter), bitsof(nd));

  QuantizationSetup();

  DWORD dwImageSize = DWORD(H1)*W2*sizeof(double),
        dwcImageSize= DWORD(h)*w2*sizeof(double),
        dwLS = (W*3/4+(((W*3)%4)?1:0))*4; // DIB line-size

  PROFILE("Init")

  // Allocating memory, we should check if ok
  lpY = (HPDOUBLE) GlobalAllocPtr(GMEM_MOVEABLE|GMEM_ZEROINIT, dwImageSize);
  lpI = (HPDOUBLE) GlobalAllocPtr(GMEM_MOVEABLE|GMEM_ZEROINIT, dwImageSize);
  lpQ = (HPDOUBLE) GlobalAllocPtr(GMEM_MOVEABLE|GMEM_ZEROINIT, dwImageSize);
  lpcY= (HPDOUBLE) GlobalAllocPtr(GMEM_MOVEABLE|GMEM_ZEROINIT, dwcImageSize);
  lpcI= (HPDOUBLE) GlobalAllocPtr(GMEM_MOVEABLE|GMEM_ZEROINIT, dwcImageSize);
  lpcQ= (HPDOUBLE) GlobalAllocPtr(GMEM_MOVEABLE|GMEM_ZEROINIT, dwcImageSize);
  if (!lpY || !lpI || !lpQ || !lpcY || !lpcI || !lpcQ) // do something
    { MsgBox("Error", "Can't allocate memory #02"); return FALSE; }

  PROFILE("Memory allocation")

  // separating color components
  ShowReset();
  ShowMessage("1 of 4 - Splitting image into YIQ color components...");
  DWORD dwSI, dwDI, dwSP, dwDP;
  for(WORD i=0; dwSI=dwLS*i, dwDI=(DWORD(i)<<Wbits), i<H; i++)
  {
    ShowRatio(float(i)/H);
    for(WORD j=0; j<W; j++)
      {
        int Y,I,Q;
        dwSP = dwSI+(j<<1)+j;
        dwDP = dwDI|DWORD(j);

        RGBtoYIQ(lpRGBbits[dwSP], lpRGBbits[dwSP+1], lpRGBbits[dwSP+2], Y, I, Q);

        lpY[dwDP]=Y;
        lpI[dwDP]=I;
        lpQ[dwDP]=Q;
      }
  }
  ShowRatio(1.0);

  PROFILE("Color separation")

  // pre-contracting color components
  ShowReset();
  ShowMessage("2 of 4 - Pre-contracting YIQ color components...");
  ContractImage(lpY, lpI, lpQ, Wbits, lpcY, lpcI, lpcQ, wbits, w,h);
  ShowRatio(1.0);

  PROFILE("Contraction")

  // Effective dencoding
  ShowReset();
  ShowMessage("4 of 4 - decoding...");
  // we should end show here, because UpdateImage call by itself Showxxxx

  double GlobalError = 0, GlobalYError=0, GlobalIError=0, GlobalQError=0;
  WORD N = B, N2=N/2;

  double * BlockY= new double[N*N];
  double * BlockI= new double[N*N];
  double * BlockQ= new double[N*N];
  BYTE * BlockRGB= new BYTE[N*N*3];


  if (img1.wWidth!=W || img1.wHeight!=H)
      {
        fprintf(rpt, "Image size missmatch: %dx%d can not map %dx%d",img1.wWidth,img1.wHeight,W,H);
        fflush(rpt);
        fclose(rpt);
        Beep();
        return FALSE;
      }

  InitImage(&img2, FALSE); // with no info
  UpdateImage(0.0);

  if (img0==NULL) img0 = &img1;  // if original image is not accessible, we take first
  double sError=0, deltarmsError=0, oldrmsError=0,
         rmsError=0, SNR, sqrSignal = sqrSum(*img0);

  if (DEC_SAVESTEPS)
    {
      char szFN[80];
      sprintf(szFN, "IFS%03u%02u.BMP", FileNameCount, 0);
      WriteDib(szFN, img2.lpDib);
      fprintf(rpt, "Image saved as %s\n",szFN); fflush(rpt);
    }

  for (WORD step=0; step<Passes; step++)
    {
    cod = fopen(szFileName,"rb");
    ReadColorHeader(cod, ID, cW, cH, B, did, djd, dRadius,
      	quantizemin, quantizemax, quantizebits);

    sError=0;

    for (WORD ir=0; ir<nr; ir++)
      {
      ShowRatio(float(step*nr+ir)/(Passes*nr));
      for (WORD jr=0; jr<mr; jr++)
        {
          int y = ir*N;
          int x = jr*N;
          WORD  idBest = y >> (bdid+1); // y/(2*did);
          WORD  jdBest = x >> (bdjd+1); // x/(2*djd);
            WORD idStart = Max(int(idBest)-dRadius, 0),
                 idEnd   = Min(int(idBest)+dRadius+1, (int)nd),
                 jdStart = Max(int(jdBest)-dRadius, 0),
                 jdEnd   = Min(int(jdBest)+dRadius+1, (int)md);
                 if (idStart==0) idEnd = Min(diameter, (int)nd);
                 if (idEnd==nd)  idStart=Max(0, int(nd)-diameter);
                 if (jdStart==0) jdEnd = Min(diameter, (int)md);
                 if (jdEnd==md)  jdStart=Max(0, int(md)-diameter);

          WORD sbb = (WORD) Unpack(4, cod);
          if (sbb!=0x0F || TRUE)
            {
            WORD id0,jd0,iso;
            DWORD ialpha,ibeta;
            double alpha,beta;

            idBest = (WORD) Unpack(idbits, cod) + idStart;
            jdBest = (WORD) Unpack(jdbits, cod) + jdStart;
            id0 = idBest*did;
            jd0 = jdBest*djd;
            iso = (WORD) Unpack(3, cod);
            if(step==1) fprintf(rpt, "%d %d, %d %d -%d-  ",
                                     idStart, idBest, jdStart, jdBest, iso);

            ialpha = (DWORD) Unpack(quantizebits[ALPHAY], cod);
            ibeta  = (DWORD) Unpack(quantizebits[BETAY] , cod);
            alpha = Dequantize(ialpha, ALPHAY);
            beta  = Dequantize(ibeta,  BETAY);
            GetBlock(lpcY, wbits, N, iso, alpha, beta, id0,jd0, BlockY);
            if(step==1) fprintf(rpt, " %03lx %03lx ", ialpha, ibeta);


            ialpha = (DWORD) Unpack(quantizebits[ALPHAI], cod);
            ibeta  = (DWORD) Unpack(quantizebits[BETAI] , cod);
            alpha = Dequantize(ialpha, ALPHAI);
            beta  = Dequantize(ibeta,  BETAI);
            GetBlock(lpcI, wbits, N, iso, alpha, beta, id0,jd0, BlockI);
            if(step==1) fprintf(rpt, " %03lx %03lx ", ialpha, ibeta);

            ialpha = (DWORD) Unpack(quantizebits[ALPHAQ], cod);
            ibeta  = (DWORD) Unpack(quantizebits[BETAQ] , cod);
            alpha = Dequantize(ialpha, ALPHAQ);
            beta  = Dequantize(ibeta,  BETAQ);
            GetBlock(lpcQ, wbits, N, iso, alpha, beta, id0,jd0, BlockQ);
            if(step==1) fprintf(rpt, " %03lx %03lx ", ialpha, ibeta);

            SetBlock(lpY, Wbits, N, y, x, BlockY);
            SetBlock(lpI, Wbits, N, y, x, BlockI);
            SetBlock(lpQ, Wbits, N, y, x, BlockQ);
            }
            if(step==1) {fprintf(rpt, "\n"); fflush(rpt);}

            for(int k=0; k<4; k++)
              if (sbb & (1<<k))
                {
                //SubBlockDecode(DS, sbp, cod, k, RBlock, Block, DBlock, Image1);
                }

            CombineYIQBlock2RGB(BlockY,BlockI,BlockQ, BlockRGB, N);
            img2.SetBlock(y, x, N, BlockRGB);
            //sError += MSDifference(*img0, ir*N, jr*N, bBlock, N);
        }
      }
    Timing = GetTimeTicks()-Time;

    sError=fabs(sError);
    rmsError = sqrt(sError/(1.0*img0->wWidth*img0->wHeight));
    deltarmsError = fabs(oldrmsError-rmsError); oldrmsError=rmsError;
    if (sError!=0 && sqrSignal!=0)
      SNR = 10*log10(sqrSignal/sError);
    else SNR = 0; // try another way to signal +infinity, -infinity
    fprintf(rpt, "Pass %d: [%lus] rms: %lf  deltarms: %lf  SNR: %lf dB\n",
                 step, (DWORD)(Timing/1000), rmsError, deltarmsError, SNR); fflush(rpt);
    fclose(cod);

    ContractImage(lpY, lpI, lpQ, Wbits, lpcY, lpcI, lpcQ, wbits, w,h);

    UpdateImage((step+1)*1.0/Passes);
    if (DEC_SAVESTEPS)
      {
        char szFN[80];
        sprintf(szFN, "IFS%03u%02u.BMP", FileNameCount, step+1);
        WriteDib(szFN, img2.lpDib);
        fprintf(rpt, "Image saved as %s\n",szFN); fflush(rpt);
      }
    if (DEC_SINGLESTEP)
      {
        Beep();
        MsgBox("","");
      }
    }// of step
    if (FirstDecode)
      { // save last step with same name as .IFS
      char szFN[80];
      strcpy(szFN, szDefName);
      strcat(szFN, ".BMP");
      WriteDib(szFN, img2.lpDib);
      fprintf(rpt, "Image saved as %s\n",szFN); fflush(rpt);
      }
    delete BlockY;
    delete BlockI;
    delete BlockQ;
    delete BlockRGB;
    GlobalFreePtr(lpY);
    GlobalFreePtr(lpI);
    GlobalFreePtr(lpQ);
    GlobalFreePtr(lpcY);
    GlobalFreePtr(lpcI);
    GlobalFreePtr(lpcQ);

    DS->rmsError=rmsError;
    DS->deltarmsError=deltarmsError;
    DS->SNR = SNR;

    fclose(rpt);
    EndImage();
    //ShowEnd();
    Beep();
    return  TRUE;
}

/*
void Kernel::SubBlockDecode(TDecStats * DS, TSubBlkParm& sbp, FILE * cod, int K, double * RBlock, double * Block, double * DBlock, double  * Image )
{
  Unpack(1, cod);  // IFS encoding
  WORD id = (WORD) Unpack(sbp.idbits, cod);
  WORD jd = (WORD) Unpack(sbp.jdbits, cod);
  int kBest = (WORD) Unpack(2, cod);
  TIsometries iso = (TIsometries) Unpack(3, cod);
  double alpha, beta;
  WORD N=B/2, W=sbp.W, H=sbp.H;

  if (sbp.DEC_QUANTIZEALPHA)
    {
    DWORD ialpha, ibeta;
    ialpha = (DWORD) Unpack(alphabits, cod);
    alpha  = double(ialpha)/((1ul<<alphabits)-1)*(alphamax-alphamin)+alphamin;
    ibeta  = (DWORD) Unpack(betabits, cod);
    beta   = double(ibeta)/((1ul<<betabits)-1)*(betamax-betamin)+betamin;
    #ifdef DEBUG_DEC_RES
//    fprintf(rpt, "R(%d,%d)<-D(%d,%d): %d %lf %lf %0lx %0lx\n",
//             ir,jr,id,jd,int(iso), alpha,beta, ialpha, ibeta); fflush(rpt);
    #endif
    }
  else
    {
    alpha = Unpackdouble(cod);
    beta  = Unpackdouble(cod);
    #ifdef DEBUG_DEC_RES
//    fprintf(rpt, "R(%d,%d)<-D(%d,%d): %d %lf %lf\n",
//             ir,jr,id,jd,int(iso), alpha,beta); fflush(rpt);
    #endif
    }

  switch (kBest)
    {
    case 0: GetBlockFromImage(Image, W, H, id*did, jd*djd, DBlock, B); break;
    case 1: GetBlockFromImage(Image, W, H, id*did, jd*djd+B, DBlock, B); break;
    case 2: GetBlockFromImage(Image, W, H, id*did+B, jd*djd, DBlock, B); break;
    case 3: GetBlockFromImage(Image, W, H, id*did+B, jd*djd+B, DBlock, B); break;
    }
  Contract(DBlock, Block, 2*N);
  BlockTransform(Block, DBlock, N, iso);
  BlockLumScale(DBlock, Block, N, alpha, beta);
  switch (K)
    {
    case 0: SetBlockToImage(RBlock, B, B, 0, 0, Block, N); break;
    case 1: SetBlockToImage(RBlock, B, B, 0, N, Block, N); break;
    case 2: SetBlockToImage(RBlock, B, B, N, 0, Block, N); break;
    case 3: SetBlockToImage(RBlock, B, B, N, N, Block, N); break;
    }
}

*/
