// file: crencode.cpp
// color Recursive encoding, 2 levels
// Written by G.Pulcini
// Last revised 1995/4/17  Rev 0.30

#include "rcencdec.h"

#define GlobalAllocPtr(x,y) malloc(y)
#define GlobalFreePtr(x) free(x)

static BYTE GetOrientation(WORD SSum[4])
{
  // ordering luminance sub-blocks 0123 index in a[] induce increasing order
  // a[0] brightest sub-block, ..., a[3] darkest sub-block (or its derivative)
  int i,k,sw,s=0,a[4]={0,1,2,3},FattTable[]={1,1,2,6,24};
  for(k=0; k<3; k++)
    {
    for(i=k+1, sw=0; i<4; i++)
      if (SSum[a[k]]<SSum[a[i]]) Swap(a[k],a[i]), sw++;
    s+=sw*FattTable[4-k-1];
    }
  return s;  // gets lexicographic order of perm.
}

void RGBLevelScale(RTImage img, int n, int d, int L)
{
  switch (L)
  {
  case 0:
    {
      for(WORD i=0; i<img.wHeight; i++)
        for (WORD j=0; j<3*img.wWidth; j++)
          img.Pixel(i,j) = img.Pixel(i,j)*n/d;
    } break;
  case 1:
    {
      for(WORD i=0; i<img.wHeight; i++)
        for (WORD j=0; j<3*img.wWidth; j++)
          img.Pixel(i,j) = 255-((255-img.Pixel(i,j))*n/d);
    } break;
  }
}

void BlockSubError(HPBYTE lpY, WORD Wbits, WORD ir0, WORD jr0, PTBLOCK B, WORD N, double E[4])
{
  E[0]=E[1]=E[2]=E[3]=0;
  WORD I, I2, N2=N/2;
  DWORD Id;
  for(WORD i=0; I=i*N, I2=i/N2*2, Id=(DWORD(ir0+i)<<Wbits)+jr0, i<N; i++)
    for(WORD j=0; j<N; j++)
      E[I2+j/N2] += sqr(double(lpY[Id+j])-double(B[I+j]));
}

void CombineYIQBlock2RGB(PTBLOCK BlockY, PTBLOCK BlockI, PTBLOCK BlockQ, PTBLOCK BlockRGB, WORD N)
{
  WORD I1,I2;
  for(WORD i=0; I2=i*N*3, I1=i*N, i<N; i++)
    for(WORD j=0; j<N; j++)
      {
        int Y,I,Q, R,G,B;
        WORD P = I2+(j<<1)+j, P1 = I1+j;
        YIQtoRGB(BlockY[P1], BlockI[P1], BlockQ[P1], R, G, B);
        BlockRGB[P]   = Bound(R, 0, 255);
        BlockRGB[P+1] = Bound(G, 0, 255);
        BlockRGB[P+2] = Bound(B, 0, 255);
      }
}

void SetBlock2RGB(PTBLOCK Block, PTBLOCK BlockRGB, WORD N)
{
  WORD I1,I2;
  for(WORD i=0; I2=i*N*3, I1=i*N, i<N; i++)
    for(WORD j=0; j<N; j++)
      {
        int Y;
        WORD P = I2+(j<<1)+j;
        Y = Block[I1+j];
        BlockRGB[P]   = Y;
        BlockRGB[P+1] = Y;
        BlockRGB[P+2] = Y;
      }
}

static void CalculateSums(HPBYTE lpcY, WORD wbits, WORD id, WORD jd, WORD B, HPDPI DPI)
// if id,jd are in range 0x0..ndxmd u should pass id*did,jd*djd to this proc
{
  DWORD I;
  WORD  B2 = B/2, I2, B4 = B/4, I4;
  WORD  Sum = 0;
  DWORD Sum2= 0;
  WORD  SSum[4]={0,0,0,0};
  DWORD SSum2[4]={0,0,0,0};
  WORD  SSSum[4][4]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  for (WORD i=0; I=DWORD(id+i)<<wbits, I2=(i/B2)<<1, I4=((i%B2)/B4)<<1, i<B; i++)
    for (WORD j=0; j<B; j++)
      {
      DWORD J = jd+j;
      WORD  SI= I2+j/B2,         // sub index
            SSI=I4+(j%B2)/B4;    // sub-sub index
      WORD xx, x = lpcY[I+J];
      Sum += x;
      Sum2+= (xx=x*x);
      SSum [SI] += x;
      SSum2[SI] += xx;
      SSSum[SI][SSI] += x;
    }
  if (DPI)
  {
    DPI->Orientation = GetOrientation(SSum);
    DPI->Sum = Sum;
    DPI->Sum2= Sum2;
    for(int k=0; k<4; k++)
      {
        DPI->SSum[k] = (SSum2[k]<<12)|SSum[k];
        DPI->SOrientation[k] = GetOrientation(SSSum[k]);
      }
  }
}

#pragma argsused
void Kernel::DoPrecalculation(BYTE  * lpDibRGB, 
        HPBYTE &lpY, HPBYTE &lpI, HPBYTE &lpQ,
        HPBYTE &lpcY,HPBYTE &lpcI,HPBYTE &lpcQ,
        HPDPI &DPI)
{}

static void Calculate(HPBYTE lpY, WORD Wbits, WORD i0, WORD j0, WORD N, WORD &Sum, DWORD &Sum2, WORD SSum[4], DWORD SSum2[4], WORD SSSum[4][4])
{
  Sum = 0;
  Sum2= 0;
  for(WORD k=0; k<4; k++)
    { SSum[k]=0;
      SSum2[k]=0;
      for(WORD k1=0; k1<4; k1++) SSSum[k][k1]=0;
    }

  DWORD I;
  WORD I2,I4, B2=N>>1, B4=N>>2;
  for(WORD i=0; I=DWORD(i0+i)<<Wbits, I2=(i/B2)<<1, I4=((i%B2)/B4)<<1, i<N; i++)
    for(WORD j=0; j<N; j++)
      {
        WORD
          SI = I2+j/B2,
          SSI= I4+(j%B2)/B4,
          xx, x = lpY[I+j0+j];
        Sum += x;
        Sum2+= (xx=x*x);
        SSum[SI] += x;
        SSum2[SI]+= xx;
        SSSum[SI][SSI] += x;
      }
}

static void BlockIsoCrossSum( HPBYTE lpY, WORD Wbits, HPBYTE lpcY, WORD wbits, WORD N, int iso,
                           WORD ir0, WORD jr0, WORD id0, WORD jd0, DWORD &SumRD)
{
  register WORD i,j;
  DWORD Ir,Id;
  SumRD=0;
  switch(iso)
  {
  case iso_IDENTITY:
     for(i=0; Ir=(DWORD(ir0+i)<<Wbits)+jr0,
              Id=(DWORD(id0+i)<<wbits)+jd0, i<N; i++)
       for(j=0; j<N; j++)
         SumRD += WORD(lpY[Ir+j])*lpcY[Id+j];
     break;
  case iso_FLIPV:
     for(i=0; Ir=(DWORD(ir0+i)<<Wbits)+jr0,
              Id=(DWORD(id0+i)<<wbits)+jd0, i<N; i++)
       for(j=0; j<N; j++)
         SumRD += WORD(lpY[Ir+j])*lpcY[Id+N-1-j];
     break;
  case iso_FLIPH:
     for(i=0; Ir=(DWORD(ir0+i)<<Wbits)+jr0,
              Id=(DWORD(id0+(N-1-i))<<wbits)+jd0, i<N; i++)
       for(j=0; j<N; j++)
         SumRD += WORD(lpY[Ir+j])*lpcY[Id+j];
     break;
  case iso_1stDIAG:
     for(i=0; Ir=(DWORD(ir0+i)<<Wbits)+jr0, i<N; i++)
       for(j=0; Id=(DWORD(id0+j)<<wbits)+jd0, j<N; j++)
         SumRD += WORD(lpY[Ir+j])*lpcY[Id+i];
     break;
  case iso_2ndDIAG:
     for(i=0; Ir=(DWORD(ir0+i)<<Wbits)+jr0, i<N; i++)
       for(j=0; j<N; j++)
         SumRD += WORD(lpY[Ir+j])*lpcY[(DWORD(id0+(N-1-j))<<wbits)+jd0+(N-1-i)];
     break;
  case iso_ROT90:
     for(i=0; Ir=(DWORD(ir0+i)<<Wbits)+jr0, i<N; i++)
       for(j=0; j<N; j++)
         SumRD += WORD(lpY[Ir+j])*lpcY[(DWORD(id0+j)<<wbits)+jd0+(N-1-i)];
     break;
  case iso_ROT180:
     for(i=0; Ir=(DWORD(ir0+i)<<Wbits)+jr0,
              Id=(DWORD(id0+(N-1-i))<<wbits)+jd0, i<N; i++)
       for(j=0; j<N; j++)
         SumRD += WORD(lpY[Ir+j])*lpcY[Id+(N-1-j)];
     break;
  case iso_ROT270:
     for(i=0; Ir=(DWORD(ir0+i)<<Wbits)+jr0, i<N; i++)
       for(j=0; j<N; j++)
         SumRD += WORD(lpY[Ir+j])*lpcY[(DWORD(id0+(N-1-j))<<wbits)+jd0+i];
     break;
  }
}

void GetBlock( HPBYTE lpcY, WORD wbits, WORD N, int iso,
                      double alpha, double beta,
                      WORD id0, WORD jd0, BYTE * Block)
{
  register WORD i,j,I;
  DWORD Id;
  switch(iso)
  {
  case iso_IDENTITY:
     for(i=0; I=i*N, Id=(DWORD(id0+i)<<wbits)+jd0, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = Bound(int(0.5+lpcY[Id+j]*alpha+beta), 0, 255);
     break;
  case iso_FLIPV:
     for(i=0; I=i*N, Id=(DWORD(id0+i)<<wbits)+jd0, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = Bound(int(0.5+lpcY[Id+N-1-j]*alpha+beta), 0, 255);
     break;
  case iso_FLIPH:
     for(i=0; I=i*N, Id=(DWORD(id0+(N-1-i))<<wbits)+jd0, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = Bound(int(0.5+lpcY[Id+j]*alpha+beta), 0, 255);
     break;
  case iso_1stDIAG:
     for(i=0; I=i*N, i<N; i++)
       for(j=0; Id=(DWORD(id0+j)<<wbits)+jd0, j<N; j++)
         Block[I+j] = Bound(int(0.5+lpcY[Id+i]*alpha+beta), 0, 255);
     break;
  case iso_2ndDIAG:
     for(i=0; I=i*N, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = Bound(int(0.5+lpcY[(DWORD(id0+(N-1-j))<<wbits)+jd0+(N-1-i)]*alpha+beta), 0, 255);
     break;
  case iso_ROT90:
     for(i=0; I=i*N, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = Bound(int(0.5+lpcY[(DWORD(id0+j)<<wbits)+jd0+(N-1-i)]*alpha+beta), 0, 255);
     break;
  case iso_ROT180:
     for(i=0; I=i*N, Id=(DWORD(id0+(N-1-i))<<wbits)+jd0, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = Bound(int(0.5+lpcY[Id+(N-1-j)]*alpha+beta), 0, 255);
     break;
  case iso_ROT270:
     for(i=0; I=i*N, i<N; i++)
       for(j=0; j<N; j++)
         Block[I+j] = Bound(int(0.5+lpcY[(DWORD(id0+(N-1-j))<<wbits)+jd0+i]*alpha+beta), 0, 255);
     break;
  }
}

void Kernel::MinimizeCRms( TEncStats &S, int ka, int kb,
                          WORD N, double SumR, double SumR2, double SumD, double SumD2, double SumRD,
                          DWORD &ialpha, DWORD &ibeta, double &alpha, double &beta, double&E )
{
  double Sum,det,oneoverSum;
  Sum   = N*N;
  oneoverSum = 1/Sum;
  det   = Sum*SumD2-SumD*SumD;

  if (det==0)
    {
      alpha = 0; S.dwDet0++;
    }
  else
    {
      alpha = (Sum*SumRD - SumR*SumD)/det;
      if (alpha>quantizemax[ka]) alpha=quantizemax[ka];
      if (alpha<quantizemin[ka]) alpha=quantizemin[ka];
    }

  ialpha = Quantize(alpha, ka);
  alpha  = Dequantize(ialpha, ka);
  beta = (SumR-alpha*SumD)*oneoverSum;

  if (beta>quantizemax[kb]) beta=quantizemax[kb];
  if (beta<quantizemin[kb]) beta=quantizemin[kb];
  ibeta = Quantize(beta, kb);
  beta  = Dequantize(ibeta, kb);
  E=(alpha*(alpha*SumD2+2*beta*SumD-2*SumRD)+beta*(beta*Sum-2*SumR)+SumR2)*oneoverSum;

  #ifdef DETAILED_RPT
  if (E<0) {fprintf(rpt, "E<0\n"); fflush(rpt); }
  #endif
}


// #define CHECKPOINT(x) MsgBox("", \"##x##\" ); // impalla il WATCOM  C++!

#define CHK(x) MsgBox("", "%d",(int)x);

#define PROFILING

#ifdef PROFILING
 #define PROFILE(x) \
  Time1 = GetTimeTicks()-Time1; \
  fprintf(rpt, x ": %lu\n", Time1); fflush(rpt); \
  Time1 = GetTimeTicks();
#elif
 #define PROFILE(x) ;
#endif

#define ShowReset() {ShowEnd(); ShowBegin();}

void Kernel::EncodeImageCR1(TEncStats &S, LPSTR szDefName, const RTImage img1, RTImage img2, TRpoolR  * RPI, TDpoolR  * DPI1, PTBLOCK DPool)
{
  ERRORCODE = 0;
  DWORD Time = GetTimeTicks();
  #ifdef PROFILING
  DWORD Time1= GetTimeTicks();
  #endif

  char szFileName[80];
  strcpy(szFileName, szDefName);
  strcat(szFileName, ".rpt");
  rpt = fopen(szFileName,"w");
  if (!rpt) { ERRORCODE = ec_noreport; return; }

  if (Weight) { delete Weight; Weight=NULL; };

  CalcScheme();

  ShowBegin();
  ShowMessage("Converting image...");
  ShowRatio(0.0);

  HPBYTE lpDibRGB = img1.lpDib;       // We should here test type of the image...

  HPBYTE lpY, lpI, lpQ,               // YIQ color components
         lpcY, lpcI, lpcQ;            // YIQ contracted color components
  HPDPI  DPI;                         // color Domain Pool Info
  HPBYTE lpRGBbits = GetDibBitsAddr(lpDibRGB);

  if ( (1u<<bitsof(djd)) != djd ||
       (1u<<bitsof(did)) != did ||
       (1u<<bitsof(B))   != B ||
        B < 2*djd ||
        B < 2*did
     ) // do something of appropriate...
     { ERRORCODE = ec_uvdims; return; }

  // uppercase are relative to image
  // lowercase are relative to contracted image

  int  dRadius = radius;  // from kernel
  int  diameter= 2*dRadius+1;
  WORD bB = bitsof(B),
       bdid = bitsof(did),
       bdjd = bitsof(djd);
  WORD W = GetDibWidth(lpDibRGB),
       H = GetDibHeight(lpDibRGB),
       Wbits = bitsof(W),
       Hbits = bitsof(H),
       W2 = WORD(1) << Wbits,
       H2 = WORD(1) << Hbits,
       mr = (W>>bB) + ((W%B)?1:0),      // W/B + ((W%B)?1:0),
       nr = (H>>bB) + ((H%B)?1:0),      // H/B + ((H%B)?1:0),
       W1 = mr<<bB,                     // mr*B,
       H1 = nr<<bB,                     // nr*B,
       w = W1>>1,
       h = H1>>1,
       wbits = bitsof(w),
       hbits = bitsof(h),
       w2 = 1u << wbits,
       h2 = 1u << hbits,
       md = (w-B)>>bdjd,                // (w-B)/djd,
       nd = (h-B)>>bdid,                // (h-B)/did,
       jdbits = Min(bitsof(diameter), bitsof(md)),
       idbits = Min(bitsof(diameter), bitsof(nd));

  QuantizationSetup();

  DWORD dwImageSize = DWORD(H1)<<Wbits,
        dwcImageSize= DWORD(h)<<wbits,
        dwLS = (W*3/4+(((W*3)%4)?1:0))*4; // DIB line-size

  PROFILE("Init")

  // Allocating memory, we should check if ok
  lpY = (LPBYTE) GlobalAllocPtr(GMEM_MOVEABLE|GMEM_ZEROINIT, dwImageSize);
  lpI = (LPBYTE) GlobalAllocPtr(GMEM_MOVEABLE|GMEM_ZEROINIT, dwImageSize);
  lpQ = (LPBYTE) GlobalAllocPtr(GMEM_MOVEABLE|GMEM_ZEROINIT, dwImageSize);
  lpcY= (LPBYTE) GlobalAllocPtr(GMEM_MOVEABLE|GMEM_ZEROINIT, dwcImageSize);
  lpcI= (LPBYTE) GlobalAllocPtr(GMEM_MOVEABLE|GMEM_ZEROINIT, dwcImageSize);
  lpcQ= (LPBYTE) GlobalAllocPtr(GMEM_MOVEABLE|GMEM_ZEROINIT, dwcImageSize);
  if (!lpY || !lpI || !lpQ || !lpcY || !lpcI || !lpcQ) 
    { ERRORCODE = ec_nomemory | 0x01; return; }

  PROFILE("Memory allocation")

  // separating color components
  ShowReset();
  ShowMessage("1 of 4 - Splitting image into YIQ color components...");

  DWORD dwSI, dwDI, dwSP, dwDP;
  for(WORD i=0; dwSI=dwLS*i, dwDI=(DWORD(i)<<Wbits), i<H; i++)
  {
    if (i & 0x0f == 0) ShowRatio(float(i)/H);
    for(WORD j=0; j<W; j++)
      {
        int R,G,B;
        int Y,I,Q;
        dwSP = dwSI+(j<<1)+j;
        dwDP = dwDI|DWORD(j);

        RGBtoYIQ(lpRGBbits[dwSP], lpRGBbits[dwSP+1], lpRGBbits[dwSP+2], Y, I, Q);

        lpY[dwDP]=Y;
        lpI[dwDP]=I;
        lpQ[dwDP]=Q;
      }
  }
  ShowRatio(1.0);

  PROFILE("Color separation")

  // pre-contracting color components
  ShowReset();
  ShowMessage("2 of 4 - Pre-contracting YIQ color components...");
  DWORD dwSI1,P0,P1,P2,P3,J;
  for(i=0; dwSI=DWORD(i<<1)<<Wbits, dwSI1=DWORD((i<<1)+1)<<Wbits, dwDI=DWORD(i)<<wbits, i<h; i++)
  {
    if (i & 0x0f == 0) ShowRatio(float(i)/h);
    for(WORD j=0; j<w; j++)
      {
        J = j<<1;
        P0 = dwSI|J;
        P1 = dwSI|(J+1);
        P2 = dwSI1|J;
        P3 = dwSI1|(J+1);
        dwDP = dwDI|DWORD(j);

        lpcY[dwDP] = 0.5+(int(lpY[P0])+int(lpY[P1])+int(lpY[P2])+int(lpY[P3]))*0.25;
        lpcI[dwDP] = 0.5+(int(lpI[P0])+int(lpI[P1])+int(lpI[P2])+int(lpI[P3]))*0.25;
        lpcQ[dwDP] = 0.5+(int(lpQ[P0])+int(lpQ[P1])+int(lpQ[P2])+int(lpQ[P3]))*0.25;
      }
  }
  ShowRatio(1.0);

  PROFILE("Contraction")

  // Now we should allocate enaugh memory to keep Domain pool
  DWORD dwDPsize = DWORD(nd)*md*sizeof(cDPoolInfo);
  DPI = (HPDPI) GlobalAllocPtr(GMEM_MOVEABLE, dwDPsize);
  if (!DPI) 
    {
      ERRORCODE = ec_nomemory | 0x02; // should deallocate ....
      return;
    }

  PROFILE("Allocating D-pool info")

  // Calculating Domain pool
  ShowReset();
  ShowMessage("3 of 4 - Pre-calculating mean and square mean of Y...");
  DWORD I;
  for (WORD id=0; I=DWORD(id)*md, id<nd; id++)
    {
    ShowRatio(float(id)/nd);
    for (WORD jd=0; jd<md; jd++)
      CalculateSums(lpcY, wbits, id<<bdid, jd<<bdjd, B, &DPI[I+jd]);
    }
  ShowRatio(1.0);

  PROFILE("Calculation of D-pool info")

  // Effective encoding
  ShowReset();
  ShowMessage("4 of 4 - Encoding...");
  // we should end show here, because UpdateImage call by itself Showxxxx

  img2.Copy(img1);
  RGBLevelScale(img2, 1,2,1);
  InitImage(&img2);
  UpdateImage(0);

  PROFILE("Image setup and level scaling")

  strcpy(szFileName, szDefName);
  strcat(szFileName, ".ifs");
  FILE * cod = fopen(szFileName,"wb");

  double GlobalError = 0, GlobalYError=0, GlobalIError=0, GlobalQError=0;
  WORD N = B, N2=N/2,
       bN=bB, bN2=bN-1;

  BYTE * BlockY= new BYTE[N*N];
  BYTE * BlockI= new BYTE[N*N];
  BYTE * BlockQ= new BYTE[N*N];
  BYTE * BlockRGB= new BYTE[N*N*3];
  if (!BlockY || !BlockI || !BlockQ || !BlockRGB) 
    { ERRORCODE = ec_nomemory | 0x03; return; }

  DWORD ID = 0x0AFFF08ul;    // stepping 08 : Color IFS !!
  DWORD PackedHeaderSize;
  // WriteColorHeader
  {
    Pack(0, 0, cod); // inits
    Pack(32, ID, cod);
    Pack(16, W, cod);
    Pack(16, H, cod);
    Pack( 8, B, cod);
    Pack( 8, did, cod);
    Pack( 8, djd, cod);
    Pack(16, dRadius, cod);
    for(int k=0; k<QCLASSES; k++)
      {
        Pack(quantizemin[k], cod);
        Pack(quantizemax[k], cod);
        Pack(8, quantizebits[k], cod);
      }
    PackedHeaderSize = Pack(-2, 0, cod);
  }

  InitEncStats(S, img1.szFileName, szDefName, W, H, nr, mr, nd, md, idbits, jdbits,
                  ID, GetRPoolInfoSize(img1), GetDPoolInfoSize(img1), GetDPoolSize(img1));
  WriteReportHeader(rpt, S,    idbits, jdbits, ID);

  PROFILE("Inits")

  double ErrorT1 = sqr(ErrorThresholdR)*B*B/4.0;
  for (WORD ir=0; ir<nr; ir++)
    {
    ShowRatio(float(ir)/nr);
    for (WORD jr=0; jr<mr; jr++)
      {
        S.dwCountBlk++;
        // This should be an else of an if which select or not poly approx.
          {
            // Full exaustive search in D-pool for best D-block
            S.dwIFSBlk++;
            TIsometries isoBest = iso_IDENTITY;
            WORD k, idBest, jdBest, y,x;
            double alpha, beta;
            DWORD ialpha, ibeta, ialphaBest, ibetaBest;
            double E,EY,EI,EQ, EBest = 10000000.0; // just a big number
            WORD  SumR, SSumR[4], SSSumR[4][4];
            DWORD SumR2, SSumR2[4];

            y = ir<<bB; // ix*N, N=B
            x = jr<<bB;
            Calculate(lpY, Wbits, y, x, N, SumR, SumR2, SSumR, SSumR2, SSSumR);
            BYTE OrientationR = GetOrientation(SSumR);

            idBest = y >> (bdid+1); // y/(2*did);
            jdBest = x >> (bdjd+1); // x/(2*djd);
            WORD idStart = Max(int(idBest)-dRadius, 0),
                 idEnd   = Min(int(idBest)+dRadius+1, (int)nd),
                 jdStart = Max(int(jdBest)-dRadius, 0),
                 jdEnd   = Min(int(jdBest)+dRadius+1, (int)md);
                 if (idStart==0) idEnd = Min(diameter, (int)nd);
                 if (idEnd==nd)  idStart=Max(0, int(nd)-diameter);
                 if (jdStart==0) jdEnd = Min(diameter, (int)md);
                 if (jdEnd==md)  jdStart=Max(0, int(md)-diameter);
            fprintf(rpt, "i::%d %d [%d] j::%d %d [%d]", idStart, idEnd, idEnd-idStart, jdStart, jdEnd, jdEnd-jdStart);
            fflush(rpt);
            DWORD I;
            for (WORD id=idStart; I=DWORD(id)*md, id<idEnd; id++) // Search in D-pool
              for (WORD jd=jdStart; jd<jdEnd; jd++)
                {
                  DWORD  IJ = I+jd;
                  WORD  SumD = DPI[IJ].Sum;
                  DWORD SumD2= DPI[IJ].Sum2;
                  DWORD SumRD;
                  TIsometries iso;

                  S.dwSearch++;
                  iso = TIsometries(
                        IsometryClass[OrientationR][DPI[IJ].Orientation]);

                  BlockIsoCrossSum(lpY, Wbits, lpcY, wbits, N, iso, y,x, id<<bdid,jd<<bdjd, SumRD);
                  MinimizeCRms( S, ALPHAY, BETAY, N, SumR, SumR2, SumD, SumD2, SumRD,
                                   ialpha, ibeta, alpha, beta, E);

                  if (E < EBest)
                    {
                      idBest = id, jdBest = jd;
                      EBest = E;
                      isoBest = iso;
                      ialphaBest = ialpha;   ibetaBest = ibeta;
                    }
                } // nested for all id,jd

            PROFILE("Full D-loop")

            WORD id0 = idBest<<bdid,  //idBest*did,
                 jd0 = jdBest<<bdjd;

            EY = fabs(EBest);
            alpha = Dequantize(ialphaBest, ALPHAY);
            beta  = Dequantize(ibetaBest,  BETAY);
            GetBlock(lpcY, wbits, N, isoBest, alpha, beta, id0,jd0, BlockY);

            // Now we are ready to evaluate the possibility of going down with recursion
            double SE[4];
            WORD sbb=0; //sub-block-bitmap
            BlockSubError(lpY, Wbits, y, x, BlockY, N, SE);
            int h = 0;
            for(k=0; k<4; k++)
                if (SE[k]>ErrorT1) { sbb |= (1<<k); h++; }
            if (h>2) { sbb = 0x0f; h=3;}
            S.wSubBlock[h]++;

            fprintf(rpt, "\t[%02x] %lf %lf %lf %lf :: %lf\n",sbb,SE[0],SE[1],SE[2],SE[3],SE[0]+SE[1]+SE[2]+SE[3]);
            fflush(rpt);

            double alphaY=alpha, betaY=beta, alphaI,alphaQ,betaI,betaQ;

            Pack(4, sbb, cod);
            if (sbb!=0x0F || TRUE)
            {
            Pack(idbits, idBest-idStart, cod);
            Pack(jdbits, jdBest-jdStart, cod);
            Pack(3, isoBest, cod);
            Pack(quantizebits[ALPHAY], ialphaBest, cod);
            Pack(quantizebits[BETAY] , ibetaBest,  cod);

            cDPoolInfo dpi;
            DWORD dwSumRD;
            double SumD,SumD2;

            BlockIsoCrossSum(lpI, Wbits, lpcI, wbits, N, isoBest, y,x, id0,jd0, dwSumRD);
            Calculate(lpI, Wbits, y, x, N, SumR, SumR2, SSumR, SSumR2, SSSumR);
            CalculateSums(lpcI, wbits, id0, jd0, B, &dpi);
            SumD = dpi.Sum;
            SumD2= dpi.Sum2;
            MinimizeCRms(S, ALPHAI, BETAI, N, SumR, SumR2, SumD, SumD2, dwSumRD,
                            ialpha, ibeta, alpha, beta, EI);
            GetBlock(lpcI, wbits, N, isoBest, alpha, beta, id0, jd0, BlockI);
            Pack(quantizebits[ALPHAI], ialpha, cod);
            Pack(quantizebits[BETAI],  ibeta,  cod);
            alphaI=alpha; betaI=beta;

            BlockIsoCrossSum(lpQ, Wbits, lpcQ, wbits, N, isoBest, y,x, id0,jd0, dwSumRD);
            Calculate(lpQ, Wbits, y, x, N, SumR, SumR2, SSumR, SSumR2, SSSumR);
            CalculateSums(lpcQ, wbits, id0, jd0, B, &dpi);
            SumD = dpi.Sum;
            SumD2= dpi.Sum2;
            MinimizeCRms(S, ALPHAQ, BETAQ, N, SumR, SumR2, SumD, SumD2, dwSumRD,
                            ialpha, ibeta, alpha, beta, EQ);
            GetBlock(lpcQ, wbits, N, isoBest, alpha, beta, id0, jd0, BlockQ);
            Pack(quantizebits[ALPHAQ], ialpha, cod);
            Pack(quantizebits[BETAQ],  ibeta,  cod);
            alphaQ=alpha; betaQ=beta;
            }

            /*
            for(k=0; k<4; k++)
              if (sbb & (1<<k))
                SubBlockEncode(S, cod, k, BlockY, ir, jr, idBest, jdBest, &(RPI[IJR]), DPI, DPool, img2, SE );
           */
            E = fabs(EY+EI+EQ)/3;
            UpdateEncStats(S, isoBest, 0, 0, E, ir, jr, idBest*2, jdBest*2 );
            fprintf(rpt, "R(%d,%d)<-D(%d,%d): %d [%lf %lf %lf] E:%lf %0.0lf%%\n",
                         ir, jr, id0*2, jd0*2, int(isoBest),
                         sqrt(EY),sqrt(EI),sqrt(EQ), sqrt(E),sqrt(E)*100/255.0);
            if (sqrt(E)>10) fprintf(rpt, "Y:%lf %lf I:%lf %lf Q:%lf %lf\n",alphaY, betaY, alphaI, betaI, alphaQ, betaQ);
            fflush(rpt);

            CombineYIQBlock2RGB(BlockY,BlockI,BlockQ, BlockRGB, N);
            //SetBlock2RGB(BlockY, BlockRGB, N);
            img2.SetBlock(y, x, N, BlockRGB);


            //GlobalError += (SE[0]+SE[1]+SE[2]+SE[3])/(B*B);
            GlobalYError+= EY;
            GlobalIError+= EI;
            GlobalQError+= EQ;
            GlobalError += E;
            UpdateImage((ir*mr+jr)/(nr*mr-1.0), ir, jr,
                         id0*2, jd0*2, int(isoBest), alpha, beta,
                         sqrt(E), sqrt(GlobalError/S.dwCountBlk));
            }
        }
      }
    delete BlockY;
    delete BlockI;
    delete BlockQ;
    delete BlockRGB;
    GlobalFreePtr(lpY);
    GlobalFreePtr(lpI);
    GlobalFreePtr(lpQ);
    GlobalFreePtr(lpcY);
    GlobalFreePtr(lpcI);
    GlobalFreePtr(lpcQ);
    GlobalFreePtr(DPI);
    Timing = GetTimeTicks()-Time;

    Pack(-1, 0, cod); // we are @ end.
    fclose(cod);

    S.packed = Pack(-2, 0, cod) - PackedHeaderSize;
    S.original=DWORD(img1.wWidth)*img1.wHeight*3;

    FixupEncStats(S, GlobalError, 0);
    WriteReportFooter(S, GlobalError, 0);

    fclose(rpt);

    EndImage();
    Beep();

  strcpy(szFileName, szDefName);
  strcat(szFileName, ".sts");
  FILE * r = fopen(szFileName,"wb");
  fwrite(&S, sizeof(TEncStats), 1, r);
  fclose(r);
}

static void GetSubBlock(int k, PTBLOCK A, WORD M, PTBLOCK B, WORD N)
/*
    0 1   values of k to get sub-block from A and put it on B
    2 3
*/
{
  WORD I,IB,i,j;
  switch(k)
  {
    case 0:
    for(i=0; I=i*N, IB=i*M, i<N; i++)
      for(j=0; j<N; j++)
        B[I+j]=A[IB+j];
    break;
    case 1:
    for(i=0; I=i*N, IB=i*M, i<N; i++)
      for(j=0; j<N; j++)
        B[I+j]=A[IB+j+N];
    break;
    case 2:
    for(i=0; I=i*N, IB=(i+N)*M, i<N; i++)
      for(j=0; j<N; j++)
        B[I+j]=A[IB+j];
    break;
    case 3:
    for(i=0; I=i*N, IB=(i+N)*M, i<N; i++)
      for(j=0; j<N; j++)
        B[I+j]=A[IB+j+N];
    break;
  }
}

void Kernel::SubBlockEncodeC( TEncStats &S, FILE * cod, int K, PTBLOCK RBlock,
                             WORD ir, WORD jr, WORD idRBest, WORD jdRBest,
                             TRpoolR far * RPI, TDpoolR  * DPI, PTBLOCK DPool,
                             RTImage img, double SE[4])
{
    double E;
    WORD N = B/2, N2=N/2;

    BYTE * Block = new BYTE[N*N];
    BYTE * DBlock= new BYTE[N*N];
    BYTE * DBlock1=new BYTE[N*N];
    BYTE * appBlock=new BYTE[N*N]; // transformed best D-Block

    WORD nd=S.nd,md=S.md;
    WORD nr=S.nr,mr=S.mr;
    WORD W = S.W,
         H = S.H;
    //optimal number of bits for Domain-(id,jd) coords
    WORD jdbits = ceil(log(md)/log(2.0)),
         idbits = ceil(log(nd)/log(2.0));

    //S.dwCountBlk++;
    //Now extract subblock from r-block
    GetSubBlock(K, RBlock, B, Block, N);
      {
        // Full exaustive search in D-pool for best D-block
        S.dwIFSBlk++;
        TIsometries isoBest = iso_IDENTITY, iso_LAST1 = iso_LAST;
        WORD k, idBest, jdBest, kBest;
        double alpha, beta, alphaBest, betaBest;
        DWORD ialpha, ibeta, ialphaBest, ibetaBest;
        double SumR = 0, SumR2 = 0,
               E0, E, E0Best, EBest = 10000000.0; // just a big number
        SumR = RPI->SubBlock[K].Sum;
        SumR2= RPI->SubBlock[K].Sum2;

        WORD idStart = Max(int(int(idRBest)-diameter+1), 0),
             idEnd   = Min(idRBest+diameter-1, nd-1),
             jdStart = Max(int(int(jdRBest)-diameter+1), 0),
             jdEnd   = Min(jdRBest+diameter-1, md-1);
             if (k=idEnd-idStart < 2*diameter-2)
             {
               if (idStart==0) idEnd = Min(idEnd+2*diameter-2-k, nd-1);
               if (idEnd==nd-1) idStart=Max(int(int(idStart)-2*diameter+2+k), 0);
             }
             if (k=jdEnd-jdStart < 2*diameter-2)
             {
               if (jdStart==0) jdEnd = Min(jdEnd+2*diameter-2-k, md-1);
               if (jdEnd==md-1) jdStart=Max(int(int(jdStart)-2*diameter+2+k), 0);
             }

        if (ENC_USEISOMETRIES && !ENC_USEORIENTATIONSCHEME) iso_LAST1 = iso_LAST;
        else iso_LAST1 = iso_IDENTITY;

        //No local search on sub-blocks by now
        //if (!ENC_OVERSTEP) overstep=1;
        //if (ENC_OVERSTEP && overstep==0) overstep=1;
        WORD I, BB=B*B, NN=N*N, md1=md/overstep, nd1=nd/overstep;
        //for (WORD id=0; I=id*md*overstep,id<nd1; id++) // Search in D-pool
        //for (WORD jd=0; jd<md1; jd++)
        //fprintf(rpt, "%d : %d..%d   %d..%d\n", diameter, idStart,idEnd, jdStart, jdEnd);
        //fflush(rpt);
        for (WORD id=idStart; I=id*md,id<idEnd; id++) // Search in D-pool
        for (WORD jd=jdStart; jd<jdEnd; jd++)
        for (WORD Sk=0; Sk<4; Sk++)
          {
            //WORD  IJ = I+jd*overstep;
            WORD IJ = I+jd;
            DWORD IJNN = ((DWORD)IJ)*BB;
            double SumD = DPI[IJ].SubBlock[Sk].Sum,
                   SumD2= DPI[IJ].SubBlock[Sk].Sum2;

            for (TIsometries iso=iso_IDENTITY; iso<=iso_LAST1; 
                iso = TIsometries(iso + TIsometries(1)) )
              {
                S.dwSearch++;
                if (ENC_USEORIENTATIONSCHEME && ENC_USEISOMETRIES)
                  {
                    iso = TIsometries(
                      IsometryClass[RPI->SubBlock[K].Orientation]
                                   [DPI[IJ].SubBlock[Sk].Orientation] );
                  }

                GetSubBlock(Sk, &DPool[IJNN], B, DBlock1, N);
                BlockTransform(DBlock1, DBlock, N, iso);

                MinimizeRms(S, Block, DBlock, N, SumR, SumR2, SumD, SumD2,
                             alpha, beta, ialpha, ibeta, E0, E);

                S.dwAlphaCount++;
                if (E < EBest)
                  {
                    idBest = id, jdBest = jd; kBest = Sk;
                    EBest = E;   E0Best = E0;
                    isoBest = iso;
                    alphaBest = alpha;     betaBest = beta;
                    ialphaBest = ialpha;   ibetaBest = ibeta;
                  }
               } // for each isometry
          } // nested for all id,jd
      /*
            // Here we have the best matching D-block with xxxxBest paremeters
            // Restore idBest, jdBest to finer step
            idBest = idBest*overstep;
            jdBest = jdBest*overstep;
            GlobalErrorwoLSS += fabs(EBest);
            TIsometries isoOldBest = isoBest;
            BOOL LSSsuccess = FALSE;
            if (ENC_OVERSTEP)
              {  // now we search D-pool with finer step around the best block found
                 // search range is -overstep+1, ..., overstep-1
              TIsometries iso_FIRST1 = iso_IDENTITY;
              if (ENC_OVERSTEPUSEISOMETRIES) iso_LAST1 = iso_LAST;
              else iso_LAST1 = isoBest, iso_FIRST1 = isoBest;

              WORD idStart = MAX(int(idBest)-overstep+1, 0),
                   idEnd   = MIN(idBest+overstep-1, md-1),
                   jdStart = MAX(int(jdBest)-overstep+1, 0),
                   jdEnd   = MIN(jdBest+overstep-1, nd-1);
              for (WORD id=idStart; I=id*md,id<=idEnd; id++) // Search in D-pool
                for (WORD jd=jdStart; jd<=jdEnd; jd++)
                {
                  if (id == idBest || jd == jdBest) continue; // skip retesting best block
                  WORD  IJ = I+jd;
                  DWORD IJNN = ((DWORD)IJ)*NN;
                  double SumD=0, SumD2=0;
                  for(k=0; k<4; k++)
                    { SumD += DPI[IJ].SubBlock[k].Sum;
                      SumD2+= DPI[IJ].SubBlock[k].Sum2;
                    }

                  for (TIsometries iso=iso_FIRST1; iso<=iso_LAST1; iso++)
                    {
                      S.dwSearch++;
                      BlockTransform(&DPool[IJNN], DBlock, N, iso);

                      MinimizeRms( rpt, Block, DBlock, N, SumR, SumR2, SumD, SumD2,
                                   alpha, beta, ialpha, ibeta, E0, E,
                                   S.dwDet0, S.dwAlphagt1, S.dwAlphalt0, S.dwAlphaltm1);
                      S.dwAlphaCount++;
                      if (E < EBest)
                        {
                          idBest = id, jdBest = jd;
                          EBest = E;   E0Best = E0;
                          isoBest = iso;
                          alphaBest = alpha;     betaBest = beta;
                          ialphaBest = ialpha;   ibetaBest = ibeta;
                          LSSsuccess=TRUE;
                        }
                    } // for each isometry
                } // for each neightbor
              }
            if (LSSsuccess) S.dwLSSsuccess++;
            if (isoOldBest==isoBest) S.dwLSSsameiso++;
      */
        SE[K] = E = fabs(EBest);
        E0Best = fabs(E0Best);
        fprintf(rpt, "\tR(%d)<-D(%d,%d,%d): %d %lf %lf [%lf] E:%lf %0.0lf%%\n",
                     K,DPI[idBest*md+jdBest].id*did,DPI[idBest*md+jdBest].jd*djd,kBest,
                     int(isoBest), alphaBest, betaBest, sqrt(E0Best), sqrt(E),sqrt(E)*100/255.0);
        fflush(rpt);

        Pack(1, 1, cod);
        Pack(idbits, DPI[idBest*md+jdBest].id, cod);
        Pack(jdbits, DPI[idBest*md+jdBest].jd, cod);
        Pack(2, kBest, cod);
        Pack(3, isoBest, cod);
        if (ENC_QUANTIZEALPHA)
          {
            Pack(alphabits, ialphaBest, cod);
            Pack(betabits, ibetaBest,  cod);
          }
        else
          {
            Pack(alphaBest,cod);
            Pack(betaBest, cod);
          }

        //UpdateEncStats(S, isoBest, alphaBest, betaBest, E, ir, jr,
        //                  DPI[idBest*md+jdBest].id, DPI[idBest*md+jdBest].jd );

        GetSubBlock(kBest, &DPool[DWORD(idBest*md+jdBest)*B*B], B, DBlock1, N);
        BlockTransform(DBlock1, DBlock, N, isoBest);
        BlockLumScale(DBlock, appBlock, N, alphaBest, betaBest);
        switch (K)
          {
          case 0: img.SetBlock(ir*B, jr*B, N, appBlock); break;
          case 1: img.SetBlock(ir*B, jr*B+N, N, appBlock); break;
          case 2: img.SetBlock(ir*B+N, jr*B, N, appBlock); break;
          case 3: img.SetBlock(ir*B+N, jr*B+N, N, appBlock); break;
          }
      }
    delete Block;
    delete DBlock;
    delete DBlock1;
    delete appBlock;
}


#undef GlobalAllocPtr
#undef GlobalFreePtr


